"""
Quantiphyse - Widgets for viewing and modifying data orientation and grids

Copyright (c) 2013-2018 University of Oxford
"""

import math

import numpy as np

try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets

from quantiphyse.gui.widgets import QpWidget, TitleWidget, NumberGrid
from quantiphyse.gui.options import OptionBox, DataOption, ChoiceOption, OutputNameOption, RunButton, NumericOption

from .processes import ResampleProcess

class ResampleDataWidget(QpWidget):
    """
    Widget that lets you resample data onto a different grid
    """
    def __init__(self, **kwargs):
        super(ResampleDataWidget, self).__init__(name="Resample Data", icon="resample.png", 
                                                 desc="Resample data onto the same grid as another data item",
                                                 group="Utilities", **kwargs)
        
    def init_ui(self):
        vbox = QtGui.QVBoxLayout()
        self.setLayout(vbox)

        vbox.addWidget(TitleWidget(self))

        optbox = OptionBox("Resampling options")
        self.data = optbox.add("Data to resample", DataOption(self.ivm))
        self.grid_data = optbox.add("Resample onto grid from", DataOption(self.ivm))
        self.order = optbox.add("Interpolation", ChoiceOption(["Nearest neighbour", "Linear", "Quadratic", "Cubic"], [0, 1, 2, 3]))
        self.output_name = optbox.add("Output name", OutputNameOption(src_data=self.data, suffix="_res"))
        vbox.addWidget(optbox)

        self.run = RunButton("Resample", self._run)
        vbox.addWidget(self.run)

        vbox.addStretch(1)
    
    def batch_options(self):
        options = {
            "data" : self.data.value,
            "grid" : self.grid_data.value,
            "output-name" : self.output_name.value,
            "order" : self.order.value,
        }

        return "Resample", options

    def _run(self):
        _, options = self.batch_options()
        ResampleProcess(self.ivm).run(options)

class OrientDataWidget(QpWidget):
    """
    Widget that lets you tweak the orientation of data
    """
    def __init__(self, **kwargs):
        super(OrientDataWidget, self).__init__(name="Orient Data", icon="inspect.png", 
                                               desc="Manipulate data orientation", 
                                               group="Utilities", **kwargs)
        self._transform_cache = {}
        self.ivm.sig_all_data.connect(self._all_data_changed)

    def init_ui(self):
        vbox = QtGui.QVBoxLayout()
        self.setLayout(vbox)

        title = TitleWidget(self)
        vbox.addWidget(title)

        hbox = QtGui.QHBoxLayout()
        self.options = OptionBox("Re-orient data")
        data = self.options.add("Data item", DataOption(self.ivm), key="data")
        data.sig_changed.connect(self._data_changed)
        self.trans, self.rot = {}, {}
        self.options.add("Translation")
        for axis, label in {2 : "axial", 0 : "sagittal", 1 : "coronal"}.items():
            trans = self.options.add("  %s (mm)" % label.title(), NumericOption(minval=-100, maxval=100, default=0), key="trans-%s" % label)
            trans.sig_changed.connect(self._translate(axis, label))
            self.trans[axis] = trans
        self.options.add("Rotation")
        for axis, label in {2 : "axial", 0 : "sagittal", 1 : "coronal"}.items():
            rot = self.options.add("  %s (degrees)" % label.title(), NumericOption(minval=-180, maxval=180, default=0), key="rot-%s" % label)
            rot.sig_changed.connect(self._rotate(axis, label))
            self.rot[axis] = rot
        hbox.addWidget(self.options)
        vbox.addLayout(hbox)

        self.gridview = GridView(self.ivm, self.ivl)
        vbox.addWidget(self.gridview)

        hbox = QtGui.QHBoxLayout()
        reset_btn = QtGui.QPushButton("Reset to original")
        reset_btn.clicked.connect(self._reset)
        hbox.addWidget(reset_btn)
        hbox.addStretch(1)
        vbox.addLayout(hbox)

        vbox.addStretch(1) 

    def activate(self):
        self._data_changed()

    def _all_data_changed(self, data):
        for name in list(self._transform_cache.keys()):
            if name not in data:
                del self._transform_cache[name]

    def _data_changed(self):
        name = self.options.values()["data"]
        qpdata = self.ivm.data.get(name, self.ivm.rois.get(name, None))
        self.gridview.set_data(qpdata)
        if qpdata is not None:
            if name not in self._transform_cache:
                self._transform_cache[name] = ([0, 0, 0], [0, 0, 0])
            translation, rotations = self._transform_cache[name]
            for axis in range(3):
                self.trans[axis].value = translation[axis]
                self.rot[axis].value = rotations[axis]
            self._set()

    def _translate(self, axis, label):
        def _trans():
            name = self.gridview.data.name
            trans = self.options.values()["trans-%s" % label]
            self._transform_cache[name][0][axis] = trans
            self._set()
        return _trans

    def _rotate(self, axis, label):
        def _rot():
            name = self.gridview.data.name
            angle = self.options.values()["rot-%s" % label]
            if axis == 1: angle = -angle
            self._transform_cache[name][1][axis] = angle
            self._set()
        return _rot

    def _reset(self):
        name = self.gridview.data.name
        del self._transform_cache[name]
        self._data_changed()

    def _set(self):
        name = self.gridview.data.name
        affine = self.gridview.data.grid.affine_orig
        grid_centre = [float(dim) / 2 for dim in self.gridview.data.grid.shape]
        world_centre = np.dot(affine[:3, :3], grid_centre)
        self.debug("Initial affine\n%s", affine)
        translation, rotations = self._transform_cache[name]
        for axis in range(3):
            angle = rotations[axis]
            rot3d = self._rotmtx_3d(axis, angle)
            affine[:3, :3] = np.dot(rot3d, affine[:3, :3])
            origin_offset = world_centre - np.dot(rot3d, world_centre)
            #origin_offset[axis] = 0
            self.debug("Origin offset\n%s", origin_offset)
            translation += origin_offset

        affine[:3, 3] += translation
        self.debug("Final affine\n%s", affine)
        self.gridview.data.grid.affine = affine
        self.gridview.update()
        # HACK
        if self.gridview.data == self.ivm.current_data:
            self.ivm.sig_current_data.emit(self.ivm.current_data)
        elif self.gridview.data == self.ivm.main:
            self.ivm.sig_main_data.emit(self.ivm.main)

    def _rotmtx_3d(self, axis, angle):
        # FIXME this is not quite right when rotating in a plane where
        # the basis vectors have different lengths
        c, s = math.cos(math.radians(angle)), math.sin(math.radians(angle))
        rot2d = np.array([[c, -s], [s, c]])
        rot3d = np.identity(3)
        if axis == 0:
            rot3d[1:, 1:] = rot2d
        elif axis == 1:
            rot3d[0, 0] = rot2d[0, 0]
            rot3d[0, 2] = rot2d[0, 1]
            rot3d[2, 0] = rot2d[1, 0]
            rot3d[2, 2] = rot2d[1, 1]
        elif axis == 2:
            rot3d[:2, :2] = rot2d
        self.debug("3d rotation matrix: %i %f", axis, angle)
        self.debug("\n%s", rot3d)
        return rot3d

class GridView(QtGui.QWidget):

    COORD_LABELS = {
        0 : "unknown",
        1 : "scanner",
        2 : "aligned",
        3 : "Talairach",
        4 : "MNI",
    }

    def __init__(self, ivm, ivl, readonly=False):
        self.ivm, self.ivl = ivm, ivl
        self.data = None
        self._updating = False

        QtGui.QWidget.__init__(self)
        grid = QtGui.QGridLayout()
        self.setLayout(grid)

        grid.addWidget(QtGui.QLabel("Grid->World Transform"), 1, 0)
        self.transform = NumberGrid(initial=np.identity(3), expandable=(False, False), fix_height=True, fix_width=True, readonly=readonly)
        self.transform.sig_changed.connect(self._changed)
        grid.addWidget(self.transform, 2, 0)

        grid.addWidget(QtGui.QLabel("Origin"), 1, 1)
        self.origin = NumberGrid(initial=[[0],]*3, expandable=(False, False), fix_height=True, fix_width=True, readonly=readonly)
        self.origin.sig_changed.connect(self._changed)
        grid.addWidget(self.origin, 2, 1)

        grid.addWidget(QtGui.QLabel("Co-ordinate system: "), 3, 0)
        self.coord_label = QtGui.QLabel("unknown")
        grid.addWidget(self.coord_label, 3, 1)

        grid.setColumnStretch(3, 1)
    
    def set_data(self, data):
        if data == self.data:
            return
        self.data = data
        self.update()

    def update(self):
        self._updating = True
        try:
            if self.data is not None:
                if hasattr(self.data, "nifti_header"):
                    self.coord_label.setText(self.COORD_LABELS[int(self.data.nifti_header['sform_code'])])
                self.transform.setValues(self.data.grid.transform)
                self.origin.setValues([[x,] for x in self.data.grid.origin])
            else:
                self.coord_label.setText("unknown")
                self.transform.setValues(np.identity(3))
                self.origin.setValues([[0],]*3)
        finally:
            self._updating = False

    def _changed(self):
        if self.data is not None and not self._updating:
            affine = self.data.grid.affine
            if self.transform.valid():
                affine[:3, :3] = self.transform.values()
            if self.origin.valid():
                affine[:3, 3] = [x[0] for x in self.origin.values()]
            self.data.grid.affine = affine
            # HACK
            if self.data == self.ivm.current_data:
                self.ivm.sig_current_data.emit(self.ivm.current_data)
            elif self.data == self.ivm.main:
                self.ivm.sig_main_data.emit(self.ivm.main)
                


"""
Quantiphyse - Picker classes for ImageView

Copyright (c) 2013-2018 University of Oxford

General picking system

On activation, a widget which wants to support picking will put the viewer in a pick mode.
Some widgets might change the mode during use (e.g. ROI builder). The viewer will clear any
existing selection at this point. Note that widgets which simply need to respond to changes
in the point of focus can connect to the ``sig_focus`` signal instead of using the picking
system.

The selected picker gets ``pick`` events from the image viewer which provide the position
picked and the source window ID. Pickers may choose to ignore pick events from windows
other than the initial source of picked points. Other pickers may accept pick events from
any window.

The viewer will also emit a selection_changed signal when the set of picked points has
changed. This might be for every click, e.g. for single/multi-point pickers, or it might
only be emitted when a particular action occurs, e.g. the user selects the last point of
a polygon, or ends a drag. Widgets may wish to connect to this event but they do not
have to. If they do connect to this event they *must* disconnect from it when not
activated, or they may receive events from different types of picker that they are not
designed to support.

Selected points are retrieved by calling the ``selected()`` method on the picker. The
objects returned are picker dependent and might be a list of point co-ordinates, a
dictionary of colour to points selected, or a 3D ROI volume. Widgets are responsible
for correctly interpreting the return value based on the picker they have chosen.

Points and ROIs are by default provided in the viewer grid space, however the optional
``grid`` parameter allows them to be provided relative to another grid.
"""

from __future__ import division, unicode_literals, absolute_import, print_function

try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets
  
import numpy as np
import pyqtgraph as pg
from PIL import Image, ImageDraw

from quantiphyse.utils import LogSource

class PickMode(object):
    """
    Enumeration of supported pick modes
    """

    #: Single point picking - see :class:`PointPicker`
    SINGLE = 1

    #: Multi-point picking - see :class:`MultiPicker`
    MULTIPLE = 2

    #: Multi-point picking in a single slice - see :class:`SliceMultiPicker`
    SLICE_MULTIPLE = 3

    #: Select rectangular regions - see :class:`RectPicker`
    RECT = 4

    #: Select elliptical regions - see :class:`EllipsePicker
    ELLIPSE = 5

    #: Select polygogn region bounded by points - see :class:`PolygonPicker`
    POLYGON = 6

    #: Select regions by dragging around them - see :class:`FreehandPicker`
    FREEHAND = 7

    #: Select a set of points by dragging - see :class:`PaintPicker`
    PAINT = 8

class Picker(LogSource):
    """
    Base class for pickers
    """
    def __init__(self, iv):
        LogSource.__init__(self)
        self.ivl = iv
        self.use_drag = False
        self.view = None
        self.win = None

    def reset(self):
        """
        Reset the picker, discarding any existing selection
        """
        pass

    def pick(self, win, pos):
        """
        Point has been picked

        :param win: ID of the window sending the event
        :param pos: Position picked
        """
        pass

    def drag(self, win, pos):
        """
        Mouse has been dragged with left button down

        :param win: ID of the window sending the event
        :param pos: Current position
        """
        pass

    def selection(self, grid=None, **kwargs):
        """
        Return the selection. This may be:

         - A single point
         - A list of points
         - A dictionary of colour to list of points
         - A region defined by a binary mask

        The return type is determined by the picker type so the caller
        must know what type of picking they are doing and what kind
        of object they expect in return

        Keyword arguments may be used by specific pickers for other purposes

        :param grid: Return point co-ordinates relative to this grid.
                     If not specified, use viewer grid
        """
        pass

    def cleanup(self):
        """
        Remove picker objects from the view
        """
        pass

class PointPicker(Picker):
    """
    Basic picker which just stores a single point
    """
    def __init__(self, ivl):
        Picker.__init__(self, ivl)
        self._point = self.ivl.focus()

    def pick(self, win, pos):
        self._point = pos

    def selection(self, grid=None, **kwargs):
        """
        :return: The selected point as a 4D co-ordinate list
        """
        if grid is not None:
            return grid.grid_to_grid(self._point, from_grid=self.ivl.grid)
        else:
            return self._point

class PaintPicker(Picker):
    """
    Picker which selects a series of points by clicking and dragging
    """
    def __init__(self, ivl):
        Picker.__init__(self, ivl)
        self._points = []
        self.use_drag = True

    def reset(self):
        self._points = []

    def pick(self, win, pos):
        self.win = win
        self.view = self.ivl.ortho_views[self.win]
        self._points = [pos,]

    def drag(self, win, pos):
        self._points.append(pos)

    def selection(self, grid=None, **kwargs):
        """
        :return: The selected point as a sequence of 4D co-ordinates
        """
        if grid is not None:
            return [grid.grid_to_grid(pos, from_grid=self.ivl.grid) for pos in self._points]
        else:
            return self._points
            
class MultiPicker(Picker):
    """
    Picker which picks multiple points, potentially in different colours
    """
    def __init__(self, ivl, col=(255, 0, 0)):
        Picker.__init__(self, ivl)
        self.col = col
        self._points = {}

    def pick(self, win, pos):
        if self.col not in self._points:
            self._points[self.col] = []
        self._points[self.col].append(pos)
        self.ivl.add_arrow(pos, col=self.col)

    def selection(self, grid=None, **kwargs):
        """
        :return: Dictionary of colour : list of picked points as 4D co-ords
        """
        if grid is not None:
            ret = {}
            for col, pts in self._points.items():
                ret[col] = [grid.grid_to_grid(pos, from_grid=self.ivl.grid) for pos in pts]
            return ret
        else:
            return self._points

    def cleanup(self):
        self.ivl.remove_arrows()

class SliceMultiPicker(MultiPicker):
    """
    MultiPicker which is restricted to a single window
    """
    def __init__(self, iv, col=(255, 0, 0)):
        MultiPicker.__init__(self, iv, col)
        self.zaxis = None
        self.zpos = None

    def pick(self, win, pos):
        if self.win is None:
            self.win = win
            self.zaxis = self.ivl.ortho_views[win].zaxis
            self.zpos = pos[self.zaxis]

        if win == self.win and pos[self.zaxis] == self.zpos:
            MultiPicker.pick(self, win, pos)

class PolygonPicker(Picker):
    """
    Picker which selects a polygon region
    """

    def __init__(self, iv):
        Picker.__init__(self, iv)
        self.roisel = None
        self._points = []

    def pick(self, win, pos):
        if self.win is None:
            self.win = win
            self.roisel = pg.PolyLineROI([], pen=(255, 0, 0))
            self.view = self.ivl.ortho_views[self.win]
            self.view.vb.addItem(self.roisel)
        elif win != self.win:
            return

        xy = self._xy_coords(pos)
        self._points.append(xy)
        self.roisel.setPoints(self._points)

    def selection(self, grid=None, **kwargs):
        """
        Get the selected points as an ROI
        """
        if self.win is None:
            return None

        if grid is None:
            grid = self.ivl.grid

        label = kwargs.get("label", 1)

        gridx, gridy, gridz, points = self._grid_points(grid)
        w, h = grid.shape[gridx], grid.shape[gridy]
        img = Image.new('L', (w, h), 0)
        ImageDraw.Draw(img).polygon(points, outline=label, fill=label)

        ret = np.zeros(grid.shape, dtype=np.int)
        slice_mask = np.array(img)
        if gridx < gridy:
            slice_mask = slice_mask.T

        slices = [slice(None)] * 3
        zpos = int(self.ivl.focus(grid)[gridz] + 0.5)
        if zpos >= 0 and zpos < grid.shape[gridz]:
            slices[gridz] = zpos
            ret[slices] = slice_mask
        return ret

    def cleanup(self):
        if self.view is not None:
            self.view.vb.removeItem(self.roisel)

    def _xy_coords(self, pos):
        return float(pos[self.view.xaxis]), float(pos[self.view.yaxis])

    def _grid_points(self, grid):
        grid_axes = grid.get_ras_axes()
        gridx = grid_axes[self.view.xaxis]
        gridy = grid_axes[self.view.yaxis]
        gridz = grid_axes[self.view.zaxis]

        # NB we are assuming here that the supplied grid is orthogonal to the
        # viewer grid, otherwise things will not work.
        self.debug("points in std space are: %s", self._points)
        points = []
        for x, y in self._points:
            std_pt = [0, 0, 0, 0]
            std_pt[self.view.xaxis] = x
            std_pt[self.view.yaxis] = y
            std_pt[self.view.zaxis] = 0
            grid_pt = grid.grid_to_grid(std_pt, from_grid=self.ivl.grid)
            points.append((int(grid_pt[gridx]+0.5), int(grid_pt[gridy]+0.5)))
        self.debug("points in grid space are: %s", points)
        return gridx, gridy, gridz, points

class RectPicker(PolygonPicker):
    """
    Picker which selects a rectangular region
    """
    def __init__(self, iv):
        PolygonPicker.__init__(self, iv)
        self.use_drag = True
        self.ox, self.oy = 0, 0

    def pick(self, win, pos):
        self.cleanup()

        self.win = win
        self.view = self.ivl.ortho_views[self.win]

        fx, fy = self._xy_coords(pos)
        self.roisel = pg.RectROI((fx, fy), (1, 1), pen=(255, 0, 255))
        self.view.vb.addItem(self.roisel)
        self._points = [(fx, fy), (fx, fy+1), (fx+1, fy+1), (fx+1, fy)]
        self.ox, self.oy = fx, fy

    def drag(self, win, pos):
        fx, fy = self._xy_coords(pos)
        sx, sy = fx-self.ox, fy-self.oy
        self.roisel.setSize((sx, sy))
        self._points = [(self.ox, self.oy), (self.ox, fy), (fx, fy), (fx, self.oy)]

class EllipsePicker(PolygonPicker):
    """
    Picker which selects an elliptical region
    """

    def __init__(self, iv):
        PolygonPicker.__init__(self, iv)
        self.use_drag = True
        self.ox, self.oy = 0, 0

    def pick(self, win, pos):
        self.cleanup()

        self.win = win
        self.view = self.ivl.ortho_views[self.win]

        fx, fy = self._xy_coords(pos)
        self.roisel = pg.EllipseROI((fx, fy), (1, 1), pen=(255, 0, 255))
        self.view.vb.addItem(self.roisel)
        self.ox, self.oy = fx, fy
        self._points = [(fx, fy), (fx, fy+1), (fx+1, fy+1), (fx+1, fy)]

    def drag(self, win, pos):
        fx, fy = self._xy_coords(pos)
        sx, sy = fx-self.ox, fy-self.oy
        if sx == 0:
            sx = 1
        if sy == 0:
            sy = 1
        self.roisel.setSize((sx, sy))
        self._points = [(self.ox, self.oy), (fx, fy)]

    def selection(self, grid=None, **kwargs):
        """
        Get the selected points as an ROI
        """
        if self.win is None:
            return None

        if grid is None:
            grid = self.ivl.grid

        label = kwargs.get("label", 1)
        
        gridx, gridy, gridz, points = self._grid_points(grid)
        # Ellipse seems to require points in ascending order for some reason
        points = [(min(points[0][0], points[1][0]), min(points[0][1], points[1][1])),
                  (max(points[0][0], points[1][0]), max(points[0][1], points[1][1]))]
        w, h = grid.shape[gridx], grid.shape[gridy]
        self.debug("getting selection with dimensions %i %i %s", w, h, label)
        img = Image.new('L', (w, h), 0)
        ImageDraw.Draw(img).ellipse(points, outline=label, fill=label)

        ret = np.zeros(grid.shape, dtype=np.int)
        slice_mask = np.array(img)
        if gridx < gridy:
            slice_mask = slice_mask.T
        self.debug("selection nonzero: %i", np.count_nonzero(slice_mask))

        slices = [slice(None)] * 3
        zpos = int(self.ivl.focus(grid)[gridz] + 0.5)
        self.debug("zpos in grid space %i", zpos)
        if zpos >= 0 and zpos < grid.shape[gridz]:
            slices[gridz] = zpos
            ret[slices] = slice_mask
        return ret

class FreehandPicker(PolygonPicker):
    """
    Picker which selects a region by dragging around the boundary
    """
    def __init__(self, iv):
        PolygonPicker.__init__(self, iv)
        self.path = None
        self.pathitem = None
        self.use_drag = True

    def pick(self, win, pos):
        self.cleanup()
        self.win = win
        self.view = self.ivl.ortho_views[self.win]

        fx, fy = float(pos[self.view.xaxis]), float(pos[self.view.yaxis])
        self._points.append((fx, fy))

        self.roisel = QtGui.QGraphicsPathItem()
        pen = QtGui.QPen(QtCore.Qt.darkMagenta, 1, QtCore.Qt.SolidLine,
                         QtCore.Qt.SquareCap, QtCore.Qt.BevelJoin)
        self.roisel.setPen(pen)
        self.view.vb.addItem(self.roisel)
        self.path = QtGui.QPainterPath(QtCore.QPointF(fx, fy))
        self.roisel.setPath(self.path)

    def drag(self, win, pos):
        if self.roisel is None:
            raise RuntimeError("roisel is None")
        else:
            fx, fy = float(pos[self.view.xaxis]), float(pos[self.view.yaxis])
            self._points.append((fx, fy))
            self.path.lineTo(fx, fy)
            self.roisel.setPath(self.path)

PICKERS = {PickMode.SINGLE : PointPicker,
           PickMode.MULTIPLE : MultiPicker,
           PickMode.SLICE_MULTIPLE : SliceMultiPicker,
           PickMode.POLYGON : PolygonPicker,
           PickMode.RECT : RectPicker,
           PickMode.ELLIPSE : EllipsePicker,
           PickMode.FREEHAND : FreehandPicker,
           PickMode.PAINT : PaintPicker,
          }



"""
Quantiphyse - Widgets which supports the writing of batch files

Copyright (c) 2013-2018 University of Oxford
"""

import os

try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets

from quantiphyse.gui.widgets import QpWidget, TitleWidget, RunBox
from quantiphyse.utils.batch import BatchScript

class BatchBuilderWidget(QpWidget):
    """
    Simple widget for building and running batch files
    """
    def __init__(self, **kwargs):
        super(BatchBuilderWidget, self).__init__(name="Batch Builder", icon="batch", 
                                                 desc="Simple helper for building and running batch files", 
                                                 group="Utilities", **kwargs)
        self.process = BatchScript(self.ivm, stdout=open(os.devnull, "w"), quit_on_exit=False)

    def init_ui(self):
        vbox = QtGui.QVBoxLayout()
        self.setLayout(vbox)

        title = TitleWidget(self, title="Batch Builder", subtitle=self.description, batch_btn=False)
        vbox.addWidget(title)

        self.proc_edit = QtGui.QPlainTextEdit()
        self.proc_edit.textChanged.connect(self._validate)
        vbox.addWidget(self.proc_edit, stretch=2)

        hbox = QtGui.QHBoxLayout()
        #self.new_case_btn = QtGui.QPushButton("New Case")
        #hbox.addWidget(self.new_case_btn)
        #hbox.setAlignment(self.new_case_btn, QtCore.Qt.AlignTop)

        self.reset_btn = QtGui.QPushButton("Reset")
        self.reset_btn.clicked.connect(self._reset)
        hbox.addWidget(self.reset_btn)
        hbox.addStretch(1)

        self.save_btn = QtGui.QPushButton("Save")
        self.save_btn.clicked.connect(self._save)
        hbox.addWidget(self.save_btn)
        vbox.addLayout(hbox)

        self.run_box = RunBox(self._get_process, self._get_options)
        vbox.addWidget(self.run_box)

        self.proc_warn = QtGui.QLabel("")
        self.proc_warn.setStyleSheet("QLabel { color : red; }")
        vbox.addWidget(self.proc_warn)

        self.default_dir = os.getcwd()
        self.changed = False
        self._reset()

    def activate(self):
        self.ivm.sig_main_data.connect(self._update)
        self.ivm.sig_all_data.connect(self._update)
        self._update()

    def deactivate(self):
        self.ivm.sig_main_data.disconnect(self._update)
        self.ivm.sig_all_data.disconnect(self._update)

    def _update(self):
        if not self.changed:
            self._reset()

    def _reset(self):
        if self.changed and self.proc_edit.toPlainText().strip() != "":
            btn = QtGui.QMessageBox.warning(self, "Reset to current data",
                                            "Changes to batch file will be lost",
                                            QtGui.QMessageBox.Ok | QtGui.QMessageBox.Cancel)
            if btn != QtGui.QMessageBox.Ok:
                return
        
        t = ""
        if self.ivm.main is not None:
            filedata = [d for d in self.ivm.data.values() if hasattr(d, "fname") and d.fname is not None]
            filerois = [d for d in self.ivm.rois.values() if hasattr(d, "fname") and d.fname is not None]
            
            casedir = os.getcwd()
            if hasattr(self.ivm.main, "fname") and self.ivm.main.fname is not None:
                casedir = os.path.dirname(self.ivm.main.fname)

            t += "OutputFolder: qp_out\n"
            t += "Debug: False\n"
            t += "\n"
            t += "Processing:\n"
            t += "  - Load:\n"
            if filedata:
                t += "      data:\n"
                for qpd in filedata:
                    d, f = os.path.split(qpd.fname)
                    if d == casedir:
                        path = f
                    else:
                        path = qpd.fname
                    t += "        %s: %s\n" % (path, qpd.name)
            if filerois:
                t += "      rois:\n"
                for qpd in filerois:
                    d, f = os.path.split(qpd.fname)
                    if d == casedir:
                        path = f
                    else:
                        path = qpd.fname
                    t += "        %s: %s\n" % (path, qpd.name)
            t += "\n"
            t += "  # Additional processing steps go here\n"
            t += "\n"
            t += "  - SaveAllExcept:\n"
            for d in filedata:
                t += "      %s:\n" % d.name
            for d in filerois:
                t += "      %s:\n" % d.name
            t += "\n"

            t += "Cases:\n"
            t += "  Case1:\n"
            t += "    Folder: %s\n" % casedir
            t += "\n"

        self.proc_edit.setPlainText(t)
        self.changed = False
        self.reset_btn.setEnabled(False)

    def _validate(self):
        self.changed = True
        self.reset_btn.setEnabled(True)
        t = self.proc_edit.toPlainText()
        self.proc_warn.setText("")
        if '\t' in self.proc_edit.toPlainText():
            self.proc_warn.setText("Tabs detected")
        else:
            try:
                self.process.run({"yaml" : t, "mode" : "check"})
            except Exception as e:
                self.proc_warn.setText("Invalid script: %s" % str(e))

    def _get_process(self):
        return self.process

    def _get_options(self):
        return {"yaml" : self.proc_edit.toPlainText()}

    def _save(self):
        fname, _ = QtGui.QFileDialog.getSaveFileName(self, 'Save batch file', 
                                                     dir=self.default_dir, 
                                                     filter="YAML files (*.yaml)")
        if fname != '':
            f = open(fname, "w")
            try:
                f.write(self.proc_edit.toPlainText())
            finally:
                f.close()
        else: # Cancelled
            pass



"""
Quantiphyse - Plotting widgets

Copyright (c) 2013-2018 University of Oxford
"""

from __future__ import division, unicode_literals, absolute_import

import numpy as np
import pyqtgraph as pg
try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets

from scipy.interpolate import UnivariateSpline

from quantiphyse.gui.widgets import OptionsButton
from quantiphyse.utils import LogSource, get_kelly_col, get_icon, norecurse

class PlotOptions(QtGui.QDialog):
    """
    Options dialog for a line plot
    """

    sig_options_changed = QtCore.Signal(object)

    def __init__(self, graph=None, **kwargs):
        QtGui.QDialog.__init__(self, graph)
        self.graph = graph
        self.sig_enh = False
        self.smooth = False

        self.setWindowTitle('Plot options')
        grid = QtGui.QGridLayout()
        self.setLayout(grid)

        row = 0
        # Display mode
        if kwargs.get("display_mode", True):
            grid.addWidget(QtGui.QLabel("Display mode"), row, 0)
            mode_combo = QtGui.QComboBox()
            mode_combo.addItem("Signal")
            mode_combo.addItem("Signal Enhancement")
            mode_combo.currentIndexChanged.connect(self._mode_changed)
            grid.addWidget(mode_combo, row, 1)
            row += 1

            # Signal enhancement baseline
            self.se_lbl = QtGui.QLabel('Signal enhancement: Use first')
            self.se_lbl.setEnabled(False)
            grid.addWidget(self.se_lbl, row, 0)

            hbox = QtGui.QHBoxLayout()
            self._norm_frames_spin = QtGui.QSpinBox()
            self._norm_frames_spin.setValue(3)
            self._norm_frames_spin.setMinimum(1)
            self._norm_frames_spin.setMaximum(100)
            self._norm_frames_spin.valueChanged.connect(self._changed)
            self._norm_frames_spin.setEnabled(False)
            hbox.addWidget(self._norm_frames_spin)
            self.se_lbl2 = QtGui.QLabel('frames as baseline')
            self.se_lbl2.setEnabled(False)
            hbox.addWidget(self.se_lbl2)
            hbox.addStretch(1)
            grid.addLayout(hbox, row, 1)
            row += 1

        # Y-axis scale
        if kwargs.get("y_scale", True):
            hbox = QtGui.QHBoxLayout()
            self.auto_y_cb = QtGui.QCheckBox('Automatic Y axis scale', self)
            self.auto_y_cb.setChecked(True)
            self.auto_y_cb.stateChanged.connect(self._auto_y_changed)
            grid.addWidget(self.auto_y_cb, row, 0)

            self.min_lbl = QtGui.QLabel("Min")
            self.min_lbl.setEnabled(False)
            hbox.addWidget(self.min_lbl)
            self.min_spin = QtGui.QDoubleSpinBox()
            self.min_spin.setMinimum(-1e20)
            self.min_spin.setMaximum(1e20)
            self.min_spin.valueChanged.connect(self._changed)
            self.min_spin.setEnabled(False)
            hbox.addWidget(self.min_spin)

            self.max_lbl = QtGui.QLabel("Max")
            self.max_lbl.setEnabled(False)
            hbox.addWidget(self.max_lbl)
            self.max_spin = QtGui.QDoubleSpinBox()
            self.max_spin.setMinimum(-1e20)
            self.max_spin.setMaximum(1e20)
            self.max_spin.valueChanged.connect(self._changed)
            self.max_spin.setEnabled(False)
            hbox.addWidget(self.max_spin)

            hbox.addStretch(1)
            grid.addLayout(hbox, row, 1)
            row += 1

        # Smoothing
        if kwargs.get("smoothing", True):
            smooth_cb = QtGui.QCheckBox('Smooth curves', self)
            smooth_cb.setChecked(self.smooth)
            smooth_cb.stateChanged.connect(self._smooth_changed)
            grid.addWidget(smooth_cb, row, 0)
            row += 1

    @property
    def yrange(self):
        return self.min_spin.value(), self.max_spin.value()

    @yrange.setter
    def yrange(self, yrange):
        try:
            self.min_spin.blockSignals(True)
            self.max_spin.blockSignals(True)
            self.min_spin.setValue(yrange[0])
            self.max_spin.setValue(yrange[1])
            self.sig_options_changed.emit(self)
        finally:
            self.min_spin.blockSignals(False)
            self.max_spin.blockSignals(False)

    @property
    def autorange(self):
        return self.auto_y_cb.isChecked()

    @autorange.setter
    def autorange(self, autorange):
        self.auto_y_cb.setChecked(autorange)

    @property
    def norm_frames(self):
        return self._norm_frames_spin.value()

    def _mode_changed(self, idx):
        self.sig_enh = (idx == 1)
        self.se_lbl.setEnabled(self.sig_enh)
        self._norm_frames_spin.setEnabled(self.sig_enh)
        self.se_lbl2.setEnabled(self.sig_enh)
        self.sig_options_changed.emit(self)

    def _auto_y_changed(self, selected):
        self.min_lbl.setEnabled(not selected)
        self.min_spin.setEnabled(not selected)
        self.max_lbl.setEnabled(not selected)
        self.max_spin.setEnabled(not selected)
        self.sig_options_changed.emit(self)
        
    def _smooth_changed(self, state):
        self.smooth = state
        self.sig_options_changed.emit(self)

    def _changed(self):
        self.sig_options_changed.emit(self)

class LinePlot(LogSource):
    """
    A 1-D array of data to be plotted as a line
    """
    def __init__(self, graph, yvalues, xvalues=None, axes=None, name=None, **kwargs):
        LogSource.__init__(self)
        if axes is None:
            self.axes = graph.axes
        if xvalues is None:
            xvalues = range(len(yvalues))

        self.name = name
        self.graph = graph
        self.axes = axes
        
        self.graph.options.sig_options_changed.connect(self._options_changed)

        self.yvalues = np.copy(np.array(yvalues, dtype=np.double))
        self.xvalues = np.copy(np.array(xvalues, dtype=np.double))
        self.line_col = kwargs.get("line_col", (255, 255, 255))
        self.line_width = kwargs.get("line_width", 1.0)
        self.line_style = kwargs.get("line_style", QtCore.Qt.SolidLine)
        self.point_col = kwargs.get("point_col", "k")
        self.point_size = kwargs.get("point_size", 5.0)
        self.point_brush = kwargs.get("point_brush", (200, 200, 200))

        self.legend_item = None
        self.graphics_items = ()
        self.show()

    def show(self):
        """
        Draw the line on to the plot
        """
        if self.graph.options.sig_enh:
            norm_frames = min(len(self.yvalues), self.graph.options.norm_frames)
            mean = np.mean(self.yvalues[:norm_frames])
            plot_values = self.yvalues / mean - 1
        else:
            plot_values = self.yvalues

        if self.graph.options.smooth:
            indexes = range(len(plot_values))
            # Tolerance does not scale by data value to scale input
            spline = UnivariateSpline(indexes, plot_values/plot_values.max(), s=0.1, k=4)
            self.debug("Number of knots in B-spline smoothing: %i", len(spline.get_knots()))
            line_values = spline(indexes)*plot_values.max()
        else:
            line_values = plot_values

        self.hide()
        # Plot line and points separately as line may be smoothed
        line = self.axes.plot(self.xvalues, line_values, 
                              pen=pg.mkPen(color=self.line_col, width=self.line_width, style=self.line_style))

        pts = self.axes.plot(self.xvalues, plot_values, 
                             pen=None, 
                             symbolBrush=self.point_brush, 
                             symbolPen=self.point_col,
                             symbolSize=self.point_size)

        self.legend_item = line
        self.graphics_items = (line, pts)

    def hide(self):
        """
        Remove the line from the plot
        """
        items = self.graphics_items
        self.graphics_items = ()
        self.legend_item = None

        for item in items:
            self.axes.removeItem(item)

    def _options_changed(self):
        if self.graphics_items:
            self.show()

class Plot(QtGui.QWidget):
    """
    Widget for plotting graphs
    """
    def __init__(self, parent=None, title="", **kwargs):
        """
        Create a plot widget

        :param qpo: Global options
        """
        QtGui.QWidget.__init__(self, parent)
        self.items = []
        self.updating = False
        self.legend_pos = (30, 30)

        vbox = QtGui.QVBoxLayout()
        self.setLayout(vbox)

        hbox = QtGui.QHBoxLayout()
        hbox.addStretch(1)

        if kwargs.get("opts_btn", True):
            self.opts_btn = OptionsButton()
            self.opts_btn.clicked.connect(self._show_options)
            hbox.addWidget(self.opts_btn)


        if kwargs.get("clear_btn", False):
            clear_icon = QtGui.QIcon(get_icon("clear"))
            self.clear_btn = QtGui.QPushButton(self)
            self.clear_btn.setIcon(clear_icon)
            self.clear_btn.setIconSize(QtCore.QSize(14, 14))
            self.clear_btn.setToolTip("Clear curves")
            self.clear_btn.clicked.connect(self.clear)
            hbox.addWidget(self.clear_btn)

        vbox.addLayout(hbox)

        self.graphics_layout = pg.GraphicsLayoutWidget()
        self.graphics_layout.setBackground(background=None)
        vbox.addWidget(self.graphics_layout, 10)

        self.axes = self.graphics_layout.addPlot()
        self.axes.setTitle(title)
        self.axes.sigRangeChanged.connect(self._range_changed)
        self.axes.vb.sigRangeChangedManually.connect(self._range_changed_manually)

        if kwargs.get("twoaxis", False):
            # For a second y-axis, create a new ViewBox, link the right axis to its coordinate system
            self.axes_alt = pg.ViewBox()
            self.axes.scene().addItem(self.axes_alt)
            self.axes.getAxis('right').linkToView(self.axes_alt)
            self.axes_alt.setXLink(self.axes)
            self.axes.vb.sigResized.connect(self._update_plot_viewbox)
        else:
            self.axes_alt = None

        self.options = PlotOptions(self, **kwargs)
        self._options_changed()
        self.options.sig_options_changed.connect(self._options_changed)
        self._regenerate_legend()

    def set_xlabel(self, name):
        """
        Set the label for the X axis
        """
        self.axes.setLabel('bottom', name)

    def set_ylabel(self, name):
        """
        Set the label for the Y axis
        """
        self.axes.setLabel('left', name)

    def add_line(self, yvalues, name=None, xvalues=None, **kwargs):
        """
        Add a line to the plot

        :param values: 1-D Array of data values
        """
        if "line_col" not in kwargs:
            kwargs["line_col"] = get_kelly_col(len(self.items))

        axes = kwargs.get("axes", "main") 
        if axes.lower() == "main":
            axes = self.axes
        elif axes.lower() == "alt":
            axes = self.axes_alt
        else:
            raise ValueError("Unknown axes: %s" % axes)

        line = LinePlot(self, yvalues, xvalues=xvalues, name=name, axes=axes, **kwargs)
        self.items.append(line)
        self._contents_changed()
        return line

    def remove(self, item):
        """
        Remove an item from the plot
        """
        item.hide()
        self.items.remove(item)
        self._contents_changed()

    def clear(self):
        """
        Clear the plot
        """
        for item in self.items:
            item.hide()
        self.items = []
        self._contents_changed()

    def _show_options(self):
        """
        Show the options dialog box
        """
        self.options.show()
        self.options.raise_()

    def _update_plot_viewbox(self):
        # Required to keep the right and left axis plots in sync with each other
        self.axes_alt.setGeometry(self.axes.vb.sceneBoundingRect())
        
        # Need to re-update linked axes since this was called
        # incorrectly while views had different shapes.
        # (probably this should be handled in ViewBox.resizeEvent)
        self.axes_alt.linkedViewChanged(self.axes.vb, self.axes_alt.XAxis)

    @norecurse
    def _range_changed(self, *args):
        self.options.yrange = self.axes.viewRange()[1]

    @norecurse
    def _range_changed_manually(self, *args):
        self.options.yrange = self.axes.viewRange()[1]
        self.options.autorange = False

    def _contents_changed(self):
        self._regenerate_legend()

    @norecurse
    def _options_changed(self, *args):
        if self.axes_alt is not None:
            self.axes_alt.enableAutoRange() # FIXME
        if self.options.autorange:
            self.axes.enableAutoRange()
        else: 
            self.axes.disableAutoRange()
            self.axes.setYRange(*self.options.yrange)

    def _regenerate_legend(self):
        # Replaces any existing legend but keep position the same in case user moved it
        # This is necessary because of what seems to be a bug in pyqtgraph which prevents
        # legend items from being removed
        if self.axes.legend:
            self.legend_pos = self.axes.legend.pos()
            if self.axes.legend.scene(): 
                self.axes.legend.scene().removeItem(self.axes.legend)
            self.axes.legend = None
    
        for item in self.items:
            if item.name:
                if not self.axes.legend:
                    self.axes.addLegend(offset=self.legend_pos)
                self.axes.legend.addItem(item.legend_item, item.name)



# -*- coding: UTF-8 -*-
__version__ = '0.9.3.post24'
__timestamp__ = 'Fri Nov 22 10:49:28 2019 +0000'
__license__ = '''Quantiphyse licence
===================

These licence terms apply to all licences granted by THE CHANCELLOR, MASTERS 
AND SCHOLARS OF THE UNIVERSITY OF OXFORD whose administrative offices are at 
office at University Offices, Wellington Square, Oxford OX1 2JD, United Kingdom
(the "University") for use of or access to Quantiphyse v0.6 ("the Software"), 
a user-friendly software for quantitative analysis of medical images. 

By using the Software provided directly to you by email, you (the "User") are 
confirming that you agree that your use of the Software is subject to these 
licence terms.

PLEASE READ THESE LICENCE TERMS CAREFULLY BEFORE USING THE SOFTWARE.  IF YOU DO 
NOT AGREE TO THESE LICENCE TERMS YOU SHOULD NOT USE THE SOFTWARE.

THE SOFTWARE ARE INTENDED FOR ACADEMICS CARRYING OUT RESEARCH AND NOT FOR USE 
BY CONSUMERS OR COMMERCIAL BUSINESSES.

1.	Academic Use Licence

1.1	The User is granted a limited non-exclusive and non-transferable royalty 
      free licence to access and use the Software provided that the User will:

  (a)	limit their use of the Software to their own internal academic 
      non-commercial research which is undertaken for the purposes of education 
      or other scholarly use; 

  (b)	not use the Software for or on behalf of any third party or to provide a 
      service or integrate all or part of the Software into a product for sale 
      or license to third parties;

  (c)	use the Software in accordance with the prevailing instructions and 
      guidance for use given on the Website and comply with procedures on the 
      Website for user identification, authentication and access;

  (d)	comply with all applicable laws and regulations with respect to their use
      of the Software; 

  (e)	except to the extent expressly permitted under these terms, not attempt 
      to: reverse compile, disassemble or copy, modify, duplicate, create 
      derivative works from, frame, mirror, republish, download, display, 
      transmit, or distribute all or any portion of the Software or Website in
      any form or media or by any means; and 

  (f)	ensure that the Copyright Notice "Copyright © 2018, University of Oxford"
      appears prominently wherever results from the Software are used, and is 
      referenced or cited with the Copyright Notice when the Software is 
      described in any research publication or on any documents or other 
      material created using the Software.

1.2	the University reserves the right at any time and without liability or 
      prior notice to the User to revise, modify and replace the functionality 
      and performance of the access to and operation of the Software. 

1.3	The User acknowledges and agrees that the University owns all 
      intellectual property rights in the Software. The User shall not have any
      right, title or interest in or to any results or other output from the 
      Software.

1.4	This Licence will terminate immediately and the User will no longer have
      any right exercise any of the rights granted to the User upon any breach
      of the conditions in Section 1.1 of this Licence.

2.	Indemnity and Liability 

2.1	The User shall defend, indemnify and hold harmless the University against
      any claims, actions, proceedings, losses, damages, expenses and costs 
      (including without limitation court costs and reasonable legal fees) 
      arising out of or in connection with the User's possession or use of the
      Software, or any breach of these terms by the User. 

2.2	The Software are provided on an 'as is' basis and the User uses the 
      Software at their own risk. No representations, conditions, warranties or
      other terms of any kind are given in respect of the Software and all 
      statutory warranties and conditions are excluded to the fullest extent 
      permitted by law. Without affecting the generality of the previous 
      sentences, the University gives no implied or express warranty and makes
      no representation that the Software or any part of them: 

  (a) will enable specific results to be obtained; or 

  (b) meets a particular specification or is comprehensive within its field or
      that it is error free or will operate without interruption; or 

  (c) is suitable for any particular, or the User's specific purposes. 

2.3	Except in relation to fraud, death or personal injury, the University's
      liability to the User for any use of the Software, in negligence or 
      arising in any other way out of the subject matter of these licence 
      terms, will not extend to any incidental or consequential damages or 
      losses, or any loss of profits, loss of revenue, loss of data, loss of
      contracts or opportunity, whether direct or indirect.

2.4	The User hereby irrevocably undertakes to the University not to make any
      claim against any employee, student, researcher or other individual 
      engaged by the University, being a claim which seeks to enforce against 
      any of them any liability whatsoever in connection with this agreement 
      or its subject-matter. 

3.	General 

3.1	Severability - If any provision (or part of a provision) of these licence
      terms is found by any court or administrative body of competent 
      jurisdiction to be invalid, unenforceable or illegal, the other 
      provisions shall remain in force.

3.2	Entire Agreement - These licence terms and any documents referred to in
      them, constitute the whole agreement between the parties and supersede 
      any previous arrangement, understanding or agreement between them 
      relating to the Software. 

3.3	Law and Jurisdiction - These licence terms and any disputes or claims 
      arising out of or in connection with them shall be governed by, and 
      construed in accordance with, the law of England. The User irrevocably 
      submits to the exclusive jurisdiction of the English courts for any 
      dispute or claim that arises out of or in connection with these licence 
      terms.

If you are interested in using the Software commercially, please contact 
Oxford University Innovation Limited to negotiate a licence. Contact details 
are chandra.ramanujan@innovation.ox.ac.uk quoting reference 14419.

# Dependencies

The following software packages are required for Quantiphyse to run and may be included in the distributed package. If you wish to modify or redistribute
these dependencies, with or without Quantiphyse, you must ensure that you comply with the terms of the relevant licenses.

## Python 2.7

The official Python license can be found at  [https://www.python.org/download/releases/2.7/license/](https://www.python.org/download/releases/2.7/license/)

## Numpy

Copyright (c) 2005, NumPy Developers

All rights reserved.

Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

-Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
-Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.
-Neither the name of the NumPy Developers nor the names of any contributors may be used to endorse or promote products derived from this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

## Pyqtgraph

Copyright (c) 2012 University of North Carolina at Chapel Hill
Luke Campagnola ('luke.campagnola@%s.com' % 'gmail')

The MIT License
Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

## Qt 

Qt 4.8 is redistributed under the [GNU Lesser Public License.](https://www.gnu.org/licenses/lgpl-2.1.en.html)

## PySide

PySide is redistributed under the [GNU Lesser Public License.](https://www.gnu.org/licenses/lgpl-2.1.en.html)

## Scipy

Copyright © 2001, 2002 Enthought, Inc.
All rights reserved.

Copyright © 2003-2013 SciPy Developers.
All rights reserved.
Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:

Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following disclaimer in the documentation and/or other materials provided with the distribution.

Neither the name of Enthought nor the names of the SciPy Developers may be used to endorse or promote products derived from this software without specific prior written permission.
THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS “AS IS” AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

## scikit-learn

New BSD License

Copyright (c) 2007–2017 The scikit-learn developers.
All rights reserved.


Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

  a. Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.
  b. Redistributions in binary form must reproduce the above copyright
     notice, this list of conditions and the following disclaimer in the
     documentation and/or other materials provided with the distribution.
  c. Neither the name of the Scikit-learn Developers  nor the names of
     its contributors may be used to endorse or promote products
     derived from this software without specific prior written
     permission. 


THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED. IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

## scikit-image

Unless otherwise specified by LICENSE.txt files in individual
directories, all code is

Copyright (C) 2011, the scikit-image team
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 3. Neither the name of skimage nor the names of its contributors may be
    used to endorse or promote products derived from this software without
    specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

## Nibabel

The MIT License

Copyright (c) 2009-2014 Matthew Brett <matthew.brett@gmail.com>
Copyright (c) 2010-2013 Stephan Gerhard <git@unidesign.ch>
Copyright (c) 2006-2014 Michael Hanke <michael.hanke@gmail.com>
Copyright (c) 2011 Christian Haselgrove <christian.haselgrove@umassmed.edu>
Copyright (c) 2010-2011 Jarrod Millman <jarrod.millman@gmail.com>
Copyright (c) 2011-2014 Yaroslav Halchenko <debian@onerussian.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

## Pynrrd

Copyright (c) 2012 Maarten H. Everts and contributors. See AUTHORS for more
details.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

## Cython

Issued under the Apache 2.0 license [https://www.apache.org/licenses/LICENSE-2.0](https://www.apache.org/licenses/LICENSE-2.0)

## Pyaml

Copyright (c) 2006 Kirill Simonov

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.

## Pillow

The Python Imaging Library (PIL) is

    Copyright © 1997-2011 by Secret Labs AB
    Copyright © 1995-2011 by Fredrik Lundh

Pillow is the friendly PIL fork. It is

    Copyright © 2010-2017 by Alex Clark and contributors

Like PIL, Pillow is licensed under the open source PIL Software License:

By obtaining, using, and/or copying this software and/or its associated documentation, you agree that you have read, understood, and will comply with the following terms and conditions:

Permission to use, copy, modify, and distribute this software and its associated documentation for any purpose and without fee is hereby granted, provided that the above copyright notice appears in all copies, and that both that copyright notice and this permission notice appear in supporting documentation, and that the name of Secret Labs AB or the author not be used in advertising or publicity pertaining to distribution of the software without specific, written prior permission.

SECRET LABS AB AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL SECRET LABS AB OR THE AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.

## Six

Copyright (c) 2010-2015 Benjamin Peterson

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
the Software, and to permit persons to whom the Software is furnished to do so,
subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

## Matplotlib

[Matplotlib license](https://matplotlib.org/users/license.html)

## Pandas

BSD 3-Clause License

Copyright (c) 2008-2012, AQR Capital Management, LLC, Lambda Foundry, Inc. and PyData Development Team
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
  list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
  this list of conditions and the following disclaimer in the documentation
  and/or other materials provided with the distribution.

* Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products derived from
  this software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

## Lmfit

[LMFIT](http://apps.jcns.fz-juelich.de/doku/sc/lmfit) is issued under the [FreeBSD license](http://opensource.org/licenses/BSD-2-Clause)

## Requests

Copyright 2017 Kenneth Reitz

Licensed under the Apache License, Version 2.0 (the "License"); you may not use this 
file except in compliance with the License. You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software distributed under 
the License is distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF 
ANY KIND, either express or implied. See the License for the specific language 
governing permissions and limitations under the License.

## Funcsigs

*funcsigs* is a derived work of CPython under the terms of the `PSF License
Agreement`. The original CPython inspect module, its unit tests and
documentation are the copyright of the Python Software Foundation. The derived
work is distributed under the `Apache License Version 2.0`

[PSF License Agreement](http://docs.python.org/3/license.html#terms-and-conditions-for-accessing-or-otherwise-using-python)
[Apache License Version 2.0](http://opensource.org/licenses/Apache-2.0)

## DEEDS 

Copyright (c) 2014, M. Heinrich
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met: 

1. Redistributions of source code must retain the above copyright notice, this
   list of conditions and the following disclaimer. 
2. Redistributions in binary form must reproduce the above copyright notice,
   this list of conditions and the following disclaimer in the documentation
   and/or other materials provided with the distribution. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

The views and conclusions contained in the software and documentation are those
of the authors and should not be interpreted as representing official policies, 
either expressed or implied, of the FreeBSD Project.

## SPHINX

Copyright (c) 2007-2019 by the Sphinx team (see AUTHORS file).
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

* Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


'''



"""
Quantiphyse - All GUI elements

Copyright (c) 2013-2018 University of Oxford
"""

from .MainWindow import MainWindow
from .ImageView import ImageView
from .ViewOptions import ViewOptions

__all__ = ["MainWindow", "ImageView", "ViewOptions",]



"""
Quantiphyse - Basic volume data classes

Concepts:

 - All data is 3D or 4D. 2D (or 1D?) data must be expanded to 3D on load. There is provision
   for 3D data to be 'interpreted' as 2D multi-volume data (there are broken Nifti files
   around like this)
 - All data is defined on a ``DataGrid`` which consists of the data shape and an affine transform
   from grid co-ordinate space to world space.
 - The 4th dimension (multiple volumes) is not part of the grid and has no impact on orientation.
 - Data objects support arbitrary slicing in a plane defined by an normal axis and a position
   relative to another grid. The output of this slice is an array, a mask (zero where outside range
   of data, plus a 2D affine transformation to map the data onto the slice grid)
 - Data objects also support resampling onto a grid, as required for some analysis methods

Copyright (c) 2013-2018 University of Oxford
"""

import logging
import math

import numpy as np
import scipy

try:
    from PySide import QtCore
except ImportError:
    from PySide2 import QtCore

from quantiphyse.utils import sf, QpException

# Private copy of pyqtgraph functions for bug fixes
from . import functions as pg

# FIXME hack to ensure extras is frozen!
from . import extras

#: Tolerance for treating values as equal
#: Used to determine if matrices are diagonal or identity
EQ_TOL = 1e-3

LOG = logging.getLogger(__name__)

def is_diagonal(mat):
    """
    :return: True if mat is diagonal, to within a tolerance of ``EQ_TOL``
    """
    return np.all(np.abs(mat - np.diag(np.diag(mat))) < EQ_TOL)

def is_identity(mat):
    """
    :return: True if mat is the identity matrix, to within a tolerance of ``EQ_TOL``
    """
    return np.all(np.abs(mat - np.identity(mat.shape[0])) < EQ_TOL)

def is_ortho_vector(vec, shape=1):
    """
    Determine if a vector has a component in only one direction

    :return: Tuple of axis, sign if vector is ortho, None, None otherwise
    """
    tol = EQ_TOL / shape
    mod = np.linalg.norm(vec)
    for axis, val in enumerate(vec):
        if abs(abs(val/mod)-1) < tol:
            return axis, math.copysign(1, val)
    return None, None

def remove_nans(x, replace_val=0):
    """
    Remove NANs from a Numpy array

    :param x: Numpy array
    :param replace_val: NaN values will be replaced with this value (default 0)
    :return: Copy of ``x`` with NaN values replaced
    """
    x = x.copy()
    x[~np.isfinite(x)] = replace_val
    return x

class DataGrid(object):
    """
    Defines a regular 3D grid in some 'world' space
    """

    def __init__(self, shape, affine, units="mm"):
        """
        Create a DataGrid object

        :param shape: Sequence of 3 integers giving number of voxels along each axis
        :param affine: 4x4 affine transformation to world co-ordinates
        """
        # Dimensionality of the grid - 3D only
        if len(shape) != 3:
            raise RuntimeError("Grid shape must be 3D")

        # 3D Affine transformation from grid-space to standard space
        # This is a 4x4 matrix - includes constant offset (origin)
        if len(affine.shape) != 2 or affine.shape[0] != 4 or affine.shape[1] != 4:
            raise RuntimeError("Grid affine must be 4x4 matrix")

        self._affine = np.copy(affine)
        self._shape = list(shape)[:]
        self._affine_orig = np.copy(affine)
        self._units = units
        
    @property
    def units(self):
        """ Units of world grid. Typically mm. Currently no support for different units in different directions """
        return self._units

    @property
    def affine(self):
        """ 4D affine matrix which describes the transformation from grid co-ordinates to world space co-ordinates"""
        return np.copy(self._affine)

    @affine.setter
    def affine(self, mat):
        self._affine = np.copy(mat)

    @property
    def affine_orig(self):
        """ Original 4D affine matrix from the initial creation of the grid. """
        return np.copy(self._affine_orig)

    @property
    def shape(self):
        """ Sequence of 3 integers giving number of voxels along each axis"""
        return np.copy(self._shape)

    @property
    def transform(self):
        """ 3x3 submatrix of ``affine`` used to transform grid space directions to world space """
        return np.copy(self._affine[:3, :3])

    @property
    def inv_transform(self):
        """ 3x3 submatrix of ``affine`` used to transform world space directions to grid space"""
        return np.linalg.inv(self.transform)

    @property
    def origin(self):
        """ 3D origin of grid in world co-ordinates (last column of ``affine``)"""
        return np.copy(self._affine[:3, 3])

    @property
    def spacing(self):
        """ Sequence of length 3 giving spacing between voxels in grid units"""
        return [np.linalg.norm(self._affine[:, i]) for i in range(3)]

    @property
    def nvoxels(self):
        """ Total number of voxels in the grid"""
        nvoxels = 1
        for dim in range(3):
            nvoxels *= self._shape[dim]
        return nvoxels

    def reset(self):
        """ Reset to original orientation """
        self.affine = self._affine_orig

    def grid_to_grid(self, coord, from_grid=None, to_grid=None, direction=False):
        """
        Transform grid co-ordinates to another grid's co-ordinates

        :param coords: 3D or 4D grid co-ordinates. If 4D, last entry is returned unchanged
        :param from_grid: DataGrid the input co-ordinates are relative to. They will be returned
                     relative to this grid
        :param to_grid: DataGrid the input co-ordinates are to be transformed into. The input
                   co-ordinates are assumed to be relative to this grid
        :return: List containing 3D or 4D co-ordinates
        """
        if from_grid is not None and to_grid is None:
            to_grid = self
        elif from_grid is None and to_grid is not None:
            from_grid = self
        else:
            raise RuntimeError("Exactly one of from_grid and to_grid must be specified")

        world = from_grid.grid_to_world(coord, direction=direction)
        return to_grid.world_to_grid(world, direction=direction)

    def grid_to_world(self, coords, direction=False):
        """
        Transform grid co-ordinates to world co-ordinates

        :param coords: 3D or 4D grid co-ordinates. If 4D, last entry is returned unchanged
        :return: List containing 3D or 4D world co-ordinates
        """
        vec = np.array(coords[:3])
        world_coords = np.dot(self.transform, vec)
        if not direction:
            world_coords += np.array(self.origin, dtype=world_coords.dtype)

        if len(coords) == 4:
            return list(world_coords) + [coords[3]]
        else:
            return list(world_coords)

    def world_to_grid(self, coords, direction=False):
        """
        Transform world co-ordinates to grid co-ordinates

        :param coords: 3D world co-ordinates. If 4D, last entry is returned unchanged
        :return: List containing 3D or 4D grid co-ordinates
        """
        vec = np.array(coords[:3])
        if not direction:
            vec -= np.array(self.origin, dtype=vec.dtype)
        grid_coords = np.dot(self.inv_transform, vec)

        if len(coords) == 4:
            return list(grid_coords) + [coords[3]]
        else:
            return list(grid_coords)

    def get_standard(self):
        """
        Return a new grid in approximate RAS order, by axis transposition
        and flipping only

        The result is a grid which defines the same space as the original
        with the same voxel spacing but where the x axis is increasing
        left->right, the y axis increases posterior->anterior and the
        z axis increases inferior->superior.
        """
        reorder, _, tmatrix = self.simplify_transforms(self.affine)
        new_shape = [self.shape[d] for d in reorder]
        return DataGrid(new_shape, tmatrix)

    def get_ras_axes(self):
        """
        Get the grid axes which best correspond to the RAS axes

        :return: List of four integers giving the axes indices of the R, A and S axes.
                 The fourth integer is always 3 indicating the volume axis
        """
        world_axes = [np.argmax(np.abs(self.transform[:, axis])) for axis in range(3)]
        grid_axes = [world_axes.index(axis) for axis in range(3)]
        return grid_axes + [3, ]

    def matches(self, grid):
        """
        Determine if another grid matches this one

        :param grid: DataGrid instance
        :return: True if ``grid`` is identical to this grid
        """
        return self.units == grid.units and np.array_equal(self.affine, grid.affine) and  np.array_equal(self.shape, grid.shape)

    def simplify_transforms(self, mat):
        """
        Convert the transformation into optional re-ordering and flipping
        of axes and a final optional affine transformation
        This enables us to use faster methods in the case where the
        grids are essentially the same or scaled
        """
        # Flip/transpose the axes to put the biggest numbers on
        # the diagonal and make negative diagonal elements positive
        dim_order, dim_flip = [], []

        absmat = np.absolute(mat)
        for dim in range(3):
            newd = np.argmax(absmat[:, dim])
            dim_order.append(newd)
            if mat[newd, dim] < 0:
                dim_flip.append(newd)

        new_mat = np.copy(mat)
        if sorted(dim_order) == range(3):
            # The transposition was consistent, so use it
            LOG.debug("Before simplification")
            LOG.debug(new_mat)
            new_shape = [self.shape[d] for d in dim_order]
            for idx, dim in enumerate(dim_order):
                new_mat[:, dim] = mat[:, idx]
            LOG.debug("After transpose %s", dim_order)
            LOG.debug(new_mat)
            for dim in dim_flip:
                # Change signs to positive to flip a dimension
                new_mat[:, dim] = -new_mat[:, dim]
            LOG.debug("After flip %s", dim_flip)
            LOG.debug(new_mat)
            for dim in dim_flip:
                # Adjust origin
                new_mat[:3, 3] = new_mat[:3, 3] - new_mat[:3, dim] * (new_shape[dim]-1)

            LOG.debug("After adjust origin %s", new_shape)
            LOG.debug(new_mat)

            return dim_order, dim_flip, new_mat
        else:
            # Transposition was inconsistent, just go with general
            # affine transform - this will work but might be slow
            return range(3), [], new_mat

class OrthoSlice(DataGrid):
    """
    DataGrid defined as a 2D orthogonal slice through another grid
    """

    def __init__(self, grid, zaxis, pos):
        """
        Create an OrthoSlice by taking a slice through an existing grid

        :param grid: DataGrid instance
        :param zaxis: Which axis is the normal axis for the slice (0, 1 or 2)
        :param pos: Position along ``axis`` to take the slice, in grid co-ordinates
        """
        affine = np.zeros((4, 4))
        shape = [1, 1, 1]
        col = 0
        for axis in range(3):
            if axis != zaxis:
                affine[:, col] = grid.affine[:, axis]
                shape[col] = grid.shape[axis]
                col += 1
            else:
                affine[:, 2] = grid.affine[:, axis]
        affine[:, 3] = grid.affine[:, 3] + pos * affine[:, 2]
        DataGrid.__init__(self, shape, affine)
        self._basis = [tuple(self.transform[:, 0]), tuple(self.transform[:, 1])]
        self._normal = tuple(self.transform[:, 2])

    @property
    def basis(self):
        """ Sequence of two 3D basis vectors for the plane in world co-ordinates"""
        return self._basis

    @property
    def normal(self):
        """ 3D normal vector to the plane in world co-ordinates"""
        return self._normal

class MetaSignaller(QtCore.QObject):
    """
    This is required because you can't multiply inherit 
    from a a QObject and a dict
    """
    sig_changed = QtCore.Signal(str)

class Metadata(dict):
    """
    Metadata dictionary.

    Emits a QT signal when keys are changed
    """
    def __init__(self, *args):
        dict.__init__(self, *args)
        #self._signaller = MetaSignaller()
        
    #@property
    #def sig_changed(self):
    #    return self._signaller.sig_changed

    def __setitem__(self, key, value):
        #self.sig_changed.emit(key)
        dict.__setitem__(self, key, value)

class QpData(object):
    """
    3D or 4D data

    Data is defined on a DataGrid instance which defines a uniform grid in standard space.

    :ivar name: Identifying name for the data the :class:`ImageVolumeManagement` class will
                require this to be a valid and Python variable name and unique within the IVM.
    :ivar grid: :class:`DataGrid` instance the data is defined on
    :ivar fname: File name data came from if relevant
    :ivar metadata: General purpose metadata dictionary. Keys are strings and values are YAML
                    convertible objects limited to the basic YAML subset used in the batch system
                    (i.e. strings, numbers and lists/dicts of these)
    """

    def __init__(self, name, grid, nvols, roi=False, metadata=None, **kwargs):
        self.name = name
        self.grid = grid

        # Number of volumes (1=3D data)
        self._nvols = nvols

        self._meta = Metadata()
        if metadata is not None:
            self._meta.update(metadata)

        self._meta["fname"] = kwargs.get("fname", None)

        # Is data set an ROI? If not specified, try making it one
        if roi is None:
            try:
                self.roi = True
            except QpException:
                self.roi = False
        else:
            self.roi = roi

        self._meta["vol_scale"] = kwargs.get("vol_scale", 1.0)
        self._meta["vol_units"] = kwargs.get("vol_units", None)

    @property
    def metadata(self):
        return self._meta

    @property
    def ndim(self):
        """ 3 or 4 for 3D or 4D data"""
        if self.nvols == 1:
            return 3
        else:
            return 4

    @property
    def nvols(self):
        """ Number of volumes (1=3D data)"""
        return self._nvols

    @property
    def shape(self):
        """ Overall data shape"""
        return list(self.grid.shape) + [self._nvols,]

    @property
    def roi(self):
        """ True if this data could be a region of interest data set"""
        return self._meta.get("roi", False)

    @roi.setter
    def roi(self, is_roi):
        if is_roi:
            if self.nvols != 1:
                raise QpException("This data set cannot be an ROI - it is 4D")
            else:
                rawdata = self.raw()
                if not np.all(np.equal(np.mod(rawdata, 1), 0)):
                    raise QpException("This data set cannot be an ROI - it does not contain integers")
        self._meta["roi"] = is_roi

    @property
    def regions(self):
        """
        Dictionary of value : name for distinct ROI region integers, not including zero
        """
        if not self.roi:
            raise TypeError("Only ROIs have distinct regions")
        
        if self._meta.get("roi_regions", None) is None:
            regions = np.unique(self.raw().astype(np.int))
            regions = np.delete(regions, np.where(regions == 0))
            if len(regions) == 1:
                # If there is only one region, don't give it a name
                self._meta["roi_regions"] = {regions[0] : ""}
            else:
                roi_regions = {}
                for region in regions:
                    roi_regions[region] = "Region %i" % region
                self._meta["roi_regions"] = roi_regions
        return self._meta["roi_regions"]

    @property 
    def fname(self):
        """
        File name origin of data or None if not from a file
        """
        return self._meta.get("fname", None)

    @fname.setter
    def fname(self, name):
        self._meta["fname"] = name
        
    def set_2dt(self):
        """
        Force 3D static data into the form of 2D multi-volume

        This is useful for some broken NIFTI files. Note that the 3D extent of the grid
        is completely ignored. In order to work, the underlying class must implement the
        change in raw().
        """
        if self.nvols != 1 or self.grid.shape[2] == 1:
            raise RuntimeError("Can only force to 2D timeseries if data was originally 3D static")

        self._meta["raw_2dt"] = True
        self._nvols = self.grid.shape[2]

        # The grid transform can't be properly interpreted because basically the file is broken,
        # so just make it 2D and hope the remaining transform is sensible
        self.grid.shape[2] = 1

    def raw(self):
        """
        Return the raw data as a Numpy array

        Concrete subclasses must implement this method to get the data from file
        or internal storage
        """
        raise NotImplementedError("Internal Error: raw() has not been implemented.")

    def volume(self, vol, qpdata=False):
        """
        Get the specified volume from a multi-volume data set

        The default implementation calls raw() to return all data. Subclasses may override
        this method, e.g. to only load the required volume.

        If the specified volume is out of range for this data, returns the last volume. Note
        that we do not copy metadata as it may be related to the whole 4D data set.

        :param vol: Volume number (0=first)
        """
        rawdata = self.raw()
        if self.ndim == 4:
            rawdata = rawdata[:, :, :, min(vol, self.nvols-1)]

        if qpdata:
            return NumpyData(rawdata, grid=self.grid, name="%s_vol_%i" % (self.name, vol))
        else:
            return rawdata

    def value(self, pos, grid=None, as_str=False):
        """
        Return the data value at a point

        :param pos: Position as a 3D or 4D vector. If 4D last value is the volume index
                    (0 for 3D). If ``grid`` not specified, position is in world space
        :param grid: If specified, interpret position in this ``DataGrid`` co-ordinate space.
        :param str: If True, return value as string to appropriate number of decimal places.
        """
        if grid is None:
            grid = DataGrid([1, 1, 1], np.identity(4))
        if len(pos) == 3:
            pos = list(pos) + [0,]
            
        data_pos = [int(math.floor(v+0.5)) for v in self.grid.grid_to_grid(pos[:3], from_grid=grid)]
        if min(data_pos) < 0:
            # Out of range but will be misinterpreted by indexing!
            value = 0
        else:
            rawdata = self.volume(pos[3])
            try:
                value = rawdata[tuple(data_pos)]
            except IndexError:
                value = 0

        if as_str:
            return sf(value)
        else:
            return value

    def timeseries(self, pos, grid=None):
        """
        Return the time/volume series at a point

        :param pos: Position as a 3D or 4D vector. If 4D last value is the volume index
                    (0 for 3D). If ``grid`` not specified, position is in world space
        :param grid: If specified, interpret position in this ``DataGrid`` co-ordinate space.
        :return: List of values, one for each volume. For 3D data, sequence has length 1.
        """
        if self.nvols == 1:
            return [self.value(pos, grid), ]

        if grid is None:
            grid = DataGrid([1, 1, 1], np.identity(4))

        data_pos = [int(math.floor(v+0.5)) for v in self.grid.grid_to_grid(pos[:3], from_grid=grid)]
        if min(data_pos) < 0:
            # Out of range but will be misinterpreted by indexing!
            return []
        else:
            rawdata = self.raw()
            try:
                return list(rawdata[data_pos[0], data_pos[1], data_pos[2], :])
            except IndexError:
                return []

    def uncache(self):
        """
        Remove large stored data arrays from memory
        
        Subclasses may implement this method to clear any caches they keep of data
        read from disk. The method will be called when the data is not active (although
        of course the data might be needed at any point). Subclasses which do not read
        data from a file might implement the method to write the data out to a temporary
        file which is then re-read on the next call to ``raw()`` or ``volume()``
        
        This method is optional and does not have to be implemented"""
        pass

    def range(self, vol=None):
        """
        Return data min and max

        Note that obtaining the range of a large 4D data set may be expensive!

        :param vol: Index of volume to use, if not specified use whole data set
        :return: Tuple of min value, max value
        """
        if vol is None:
            if self._meta.get("range", None) is None:
                self._meta["range"] = np.nanmin(self.raw()), np.nanmax(self.raw())
            return self._meta["range"]
        else:
            voldata = self.volume(vol)
            return np.nanmin(voldata), np.nanmax(voldata)

    def mask(self, roi, region=None, vol=None, invert=False, output_flat=False, output_mask=False):
        """
        Mask the data

        :param roi: ROI data item If None, return all data
        :param region: If specified, return data within this region. Otherwise return
                       data within any ROI region.
        :param vol: If specified, restrict output to this volume index
        :param invert: If True, invert the mask
        :param flat: If True, return unmasked data only as flattened arrray
        :param mask: If True, return boolean mask in same grid space as data
        :return If ``flat``, 1-D Numpy array containing unmasked data. Otherwise, Numpy array
                containing data with masked points zeroed. If ``mask`` also return
                a boolean mask in the same grid space as the data.
        """
        ret = []
        if vol is not None:
            data = self.volume(vol)
        else:
            data = self.raw()

        if roi is None:
            if output_flat:
                ret.append(data.flatten())
            else:
                ret.append(data)
            if output_mask:
                ret.append(np.ones(data.shape[:3], dtype=np.int))
        else:
            roi = roi.resample(self.grid)
            if region is None:
                mask = roi.raw() > 0
            else:
                mask = roi.raw() == region
            if invert:
                mask = np.logical_not(mask)

            if output_flat:
                ret.append(data[mask])
            else:
                masked = np.zeros(data.shape)
                masked[mask] = data[mask]
                ret.append(masked)
            if output_mask:
                ret.append(mask)

        if len(ret) > 1:
            return tuple(ret)
        else:
            return ret[0]

    def resample(self, grid, order=0, suffix="_resampled"):
        """
        Resample the data onto a new grid

        :param grid: :class:`DataGrid` to resample the data on to
        :return: New :class:`QpData` object
        """
        data = self.raw()

        LOG.debug("Resampling from:")
        LOG.debug(self.grid.affine)
        LOG.debug("To:")
        LOG.debug(grid.affine)

        # Affine transformation matrix from current grid to new grid
        tmatrix = np.dot(np.linalg.inv(grid.affine), self.grid.affine)
        reorder, flip, tmatrix = self.grid.simplify_transforms(tmatrix)

        # Perform the flips and transpositions which simplify the transformation and
        # may avoid the need for an affine transformation, or make it a simple scaling
        if reorder != range(3):
            if data.ndim == 4:
                reorder = reorder + [3]
            data = np.transpose(data, reorder)
            
        if flip:
            for dim in flip:
                data = np.flip(data, dim)

        if not is_identity(tmatrix):
            # We were not able to reduce the transformation down to flips/transpositions
            # so we will need an affine transformation
            # scipy requires the out->in transform so invert our in->out transform
            tmatrix = np.linalg.inv(tmatrix)
            affine = tmatrix[:3, :3]
            offset = list(tmatrix[:3, 3])
            output_shape = list(grid.shape[:])
            if data.ndim == 4:
                # Make 4D affine with identity transform in 4th dimension
                affine = np.append(affine, [[0, 0, 0]], 0)
                affine = np.append(affine, [[0], [0], [0], [1]], 1)
                offset.append(0)
                output_shape.append(data.shape[3])

            if is_diagonal(affine):
                # The transformation is diagonal, so use faster sequence mode
                affine = np.diagonal(affine)
            #LOG.debug("WARNING: affine_transform: ")
            #LOG.debug(affine)
            #LOG.debug("Offset = ", offset)
            #LOG.debug("Input shape=", data.shape, data.min(), data.max())
            #LOG.debug("Output shape=", output_shape)
            data = scipy.ndimage.affine_transform(data, affine, offset=offset,
                                                  output_shape=output_shape, order=order)

            if self.roi:
                # If source data was ROI, output should be, however resampling could have
                # led to non-integer data
                data = data.astype(np.int32)

        return NumpyData(data=data, grid=grid, name=self.name + suffix, roi=self.roi, metadata=self._meta)

    def slice_data(self, plane, vol=0, interp_order=0):
        """
        Extract a data slice in raw data resolution

        :param plane: OrthoSlice representing the slice to be extracted. Note that this
                      slice will not in general be defined on the same grid as the data
        :param vol: volume index for use if this is a 4D data set
        :param interp_order: Order of interpolation for non-orthogonal slices
        """
        rawdata = self.volume(vol)
        
        data_origin = np.array(self.grid.grid_to_grid([0, 0, 0], from_grid=plane))
        data_normal = np.array(self.grid.grid_to_grid([0, 0, 1], from_grid=plane, direction=True))

        data_scaled_normal = np.array([data_normal[idx] * self.grid.spacing[idx] * self.grid.spacing[idx] for idx in range(3)])
        data_naxis = np.argmax(np.absolute(data_scaled_normal))

        data_axes = list(range(3))
        del data_axes[data_naxis]
        slice_basis, slice_shape = [], []
        for axis in data_axes:
            vec = [0, 0, 0]
            vec[axis] = 1
            vec[data_naxis] = -(data_scaled_normal[axis] / data_scaled_normal[data_naxis])
            slice_basis.append(vec)
            slice_shape.append(rawdata.shape[axis])

        trans_v = np.array([
            np.array(self.grid.grid_to_grid(slice_basis[0], to_grid=plane, direction=True)),
            np.array(self.grid.grid_to_grid(slice_basis[1], to_grid=plane, direction=True)),
        ])

        trans_v = np.delete(trans_v, 2, 1)
        slice_origin = [0, 0, 0]
        slice_origin[data_naxis] = np.dot(data_origin, data_scaled_normal) / data_scaled_normal[data_naxis]
        data_offset = data_origin - slice_origin + 0.5
        offset = -np.array(self.grid.grid_to_grid(data_offset, to_grid=plane, direction=True))[:2]

        ax1, sign1 = is_ortho_vector(slice_basis[0], slice_shape[0])
        ax2, sign2 = is_ortho_vector(slice_basis[1], slice_shape[0])
        if ax1 is not None and ax2 is not None:
            #LOG.debug("\nOrthoSlice: data basis: %s (shape=%s)" % (str(slice_basis), str(slice_shape)))
            slices = [None, None, None]
            slices[ax1] = self._get_slice(rawdata.shape[ax1], sign1)
            slices[ax2] = self._get_slice(rawdata.shape[ax2], sign2)

            pos = int(math.floor(data_origin[data_naxis]+0.5))
            if pos >= 0 and pos < rawdata.shape[data_naxis]:
                slices[data_naxis] = pos
                LOG.debug("Using Numpy slice: %s %s", slices, rawdata.shape)
                sdata = rawdata[slices]
                smask = np.ones(slice_shape)
            else:
                # Requested slice is outside the data range
                LOG.debug("Outside data range: %i, %i", pos, rawdata.shape[data_naxis])
                sdata = np.zeros(slice_shape)
                smask = np.zeros(slice_shape)
        else:
            LOG.debug("Full affine slice")

            #LOG.debug("OrthoSlice: plane origin: %s" % str(plane.origin))
            #LOG.debug("OrthoSlice: plane v1: %s" % str(plane.basis[0]))
            #LOG.debug("OrthoSlice: plane v2: %s" % str(plane.basis[1]))
            #LOG.debug("OrthoSlice: plane n: %s" % str(plane.normal))
            #LOG.debug("OrthoSlice: Plane affine")
            #LOG.debug(plane.affine)

            #LOG.debug("OrthoSlice: data origin: %s" % str(data_origin))
            #LOG.debug("OrthoSlice: data n: %s" % str(data_normal))
            #LOG.debug("OrthoSlice: data naxis: %i" % data_naxis)
            #LOG.debug("OrthoSlice: data affine")
            #LOG.debug(self.grid.affine)
            
            #LOG.debug("OrthoSlice: new normal=", data_scaled_normal)
            #LOG.debug("OrthoSlice: data basis: %s (shape=%s)" % (str(slice_basis), str(slice_shape)))
            #LOG.debug("OrthoSlice: trans matrix:\n%s" % (str(trans_v)))
            #LOG.debug("OrthoSlice: slice origin=", slice_origin)
            #LOG.debug("OrthoSlice: new origin: %s (offset=%s)" % (str(slice_origin), str(data_offset)))
            #LOG.debug("OrthoSlice: new plane offset: %s" % (str(offset)))
            
            #LOG.debug("Origin: ", slice_origin)
            #LOG.debug("Basis", slice_basis)
            #LOG.debug("Shape", slice_shape)
            if self.roi:
                # Use nearest neighbour interpolation for ROIs
                sdata = pg.affineSlice(rawdata, slice_shape, slice_origin, slice_basis, range(3), order=0)
                smask = np.ones(sdata.shape)
            else:
                # Generate mask by flagging out of range data with value less than data minimum
                dmin = np.min(rawdata)
                sdata = pg.affineSlice(rawdata, slice_shape, slice_origin, slice_basis, range(3), 
                                       order=interp_order, mode='constant', cval=dmin-100)
                smask = np.ones(sdata.shape)
                smask[sdata < dmin] = 0
                sdata[sdata < dmin] = 0
            LOG.debug("Done affine slice")

        return remove_nans(sdata), smask, trans_v, offset

    def _get_slice(self, length, sign):
        if sign == 1:
            return slice(0, length, 1)
        else:
            return slice(length-1, None, -1)

    def get_bounding_box(self, ndim=3):
        """
        Returns a sequence of slice objects which
        describe the bounding box of this ROI.
        If ndim is specified, will return a bounding
        box of this number of dimensions, truncating
        and appending slices as required.

        This enables data to be
        easily restricted to the ROI region and
        reduce data copying.

        e.g.
        slices = roi.get_bounding_box(img.ndim)
        img_restric = data[slices]
        ... process img_restict, returning out_restrict
        out_full = np.zeros(img.shape)
        out_full[slices] = out_restrict
        """
        if not self.roi:
            raise RuntimeError("get_bounding_box() called on non-ROI data")

        slices = [slice(None)] * ndim
        for dim in range(min(ndim, 3)):
            axes = [i for i in range(3) if i != dim]
            nonzero = np.any(self.raw(), axis=tuple(axes))
            bb_start, bb_end = np.where(nonzero)[0][[0, -1]]
            slices[dim] = slice(bb_start, bb_end+1)

        return slices

class NumpyData(QpData):
    """
    QpData instance with in-memory Numpy data
    """
    def __init__(self, data, grid, name, **kwargs):
        # Unlikely but possible that first data is added from the console. In this
        # case no grid will exist
        if grid is None:
            grid = DataGrid(data.shape[:3], np.identity(4))

        if data.dtype.kind in np.typecodes["AllFloat"]:
            # Use float32 rather than default float64 to reduce storage
            data = data.astype(np.float32)
        self.rawdata = data
        
        if data.ndim > 3:
            nvols = data.shape[3]
            if nvols == 1:
                self.rawdata = np.squeeze(self.rawdata, axis=-1)
        else:
            nvols = 1

        QpData.__init__(self, name, grid, nvols, **kwargs)
    
    def raw(self):
        if self._meta.get("raw_2dt", False) and self.rawdata.ndim == 3:
            # Single-slice, interpret 3rd dimension as time
            return np.expand_dims(self.rawdata, 2)
        else:
            return self.rawdata



"""
Quantiphyse - Generic analysis processes

Copyright (c) 2013-2018 University of Oxford
"""
import sys

import six
import numpy as np
import scipy

try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets

from quantiphyse.data import NumpyData, OrthoSlice
from quantiphyse.utils import QpException, table_to_extra, sf
from quantiphyse.processes import Process

class CalcVolumesProcess(Process):
    """
    Calculate volume of ROI region or regions
    """

    PROCESS_NAME = "CalcVolumes"

    def __init__(self, ivm, **kwargs):
        Process.__init__(self, ivm, **kwargs)
        self.model = QtGui.QStandardItemModel()

    def run(self, options):
        self.model.clear()
        self.model.setVerticalHeaderItem(0, QtGui.QStandardItem("Num voxels"))
        self.model.setVerticalHeaderItem(1, QtGui.QStandardItem("Volume (mm^3)"))

        roi_name = options.pop('roi', None)
        sel_region = options.pop('region', None)

        if roi_name is None:
            roi = self.ivm.current_roi
        else:
            roi = self.ivm.rois[roi_name]

        if roi is not None:
            sizes = roi.grid.spacing
            counts = np.bincount(roi.raw().flatten().astype(np.int))
            col_idx = 0
            for region, name in roi.regions.items():
                if sel_region is None or region == sel_region:
                    nvoxels = counts[region]
                    vol = counts[region]*sizes[0]*sizes[1]*sizes[2]
                    self.model.setHorizontalHeaderItem(col_idx, QtGui.QStandardItem(name))
                    self.model.setItem(0, col_idx, QtGui.QStandardItem(str(nvoxels)))
                    self.model.setItem(1, col_idx, QtGui.QStandardItem(str(vol)))
                    col_idx += 1

        if not options.pop('no-extras', False):
            output_name = options.pop('output-name', "roi-vols")
            self.ivm.add_extra(output_name, table_to_extra(self.model, output_name))

class DataStatisticsProcess(Process):
    """
    Calculate summary statistics on data
    """
    
    PROCESS_NAME = "DataStatistics"
    
    def __init__(self, ivm, **kwargs):
        Process.__init__(self, ivm, **kwargs)
        self.model = QtGui.QStandardItemModel()

    def run(self, options):
        data_name = options.pop('data', None)
        output_name = options.pop('output-name', None)
        if data_name is None:
            data_items = self.ivm.data.keys()
        elif isinstance(data_name, six.string_types):
            data_items = [data_name,]
            if output_name is None:
                output_name = "%s_stats" % data_name
        else:
            data_items = data_name
        data_items = [self.ivm.data[name] for name in data_items]
        if output_name is None:
            output_name = "stats"

        roi_name = options.pop('roi', None)
        roi = None
        if roi_name is not None:
            roi = self.ivm.rois[roi_name]

        slice_dir = options.pop('slice-dir', None)
        slice_pos = options.pop('slice-pos', 0)
        sl = None
        if slice_dir is not None:
            sl = OrthoSlice(self.ivm.main.grid, slice_dir, slice_pos)
        
        vol = options.pop('vol', None)

        no_extra = options.pop('no-extras', False)
        hist_bins = options.pop('hist-bins', 20)
        hist_range = options.pop('hist-bins', None)
        
        self.model.clear()
        self.model.setVerticalHeaderItem(0, QtGui.QStandardItem("Mean"))
        self.model.setVerticalHeaderItem(1, QtGui.QStandardItem("Median"))
        self.model.setVerticalHeaderItem(2, QtGui.QStandardItem("STD"))
        self.model.setVerticalHeaderItem(3, QtGui.QStandardItem("Min"))
        self.model.setVerticalHeaderItem(4, QtGui.QStandardItem("Max"))

        col = 0
        for data in data_items:
            stats1, roi_labels, _, _ = self.get_summary_stats(data, roi, hist_bins=hist_bins, hist_range=hist_range, slice_loc=sl, vol=vol)
            for ii in range(len(stats1['mean'])):
                self.model.setHorizontalHeaderItem(col, QtGui.QStandardItem("%s\n%s" % (data.name, roi_labels[ii])))
                self.model.setItem(0, col, QtGui.QStandardItem(sf(stats1['mean'][ii])))
                self.model.setItem(1, col, QtGui.QStandardItem(sf(stats1['median'][ii])))
                self.model.setItem(2, col, QtGui.QStandardItem(sf(stats1['std'][ii])))
                self.model.setItem(3, col, QtGui.QStandardItem(sf(stats1['min'][ii])))
                self.model.setItem(4, col, QtGui.QStandardItem(sf(stats1['max'][ii])))
                col += 1

        if not no_extra: 
            self.ivm.add_extra(output_name, table_to_extra(self.model, output_name))

    def get_summary_stats(self, data, roi=None, hist_bins=20, hist_range=None, slice_loc=None, vol=None):
        """
        Get summary statistics

        :param data: QpData instance for the data to get stats from
        :param roi: Restrict data to within this roi

        :return: Sequence of summary stats dictionary, roi labels
        """
        # Checks if either ROI or data is None
        if roi is not None:
            roi = roi.resample(data.grid)
        else:
            roi = NumpyData(np.ones(data.grid.shape[:3]), data.grid, "temp", roi=True)

        if data is None:
            stat1 = {'mean': [0], 'median': [0], 'std': [0], 'max': [0], 'min': [0]}
            return stat1, list(roi.regions.keys()), np.array([0, 0]), np.array([0, 1])

        stat1 = {'mean': [], 'median': [], 'std': [], 'max': [], 'min': []}
        hist1 = []
        hist1x = []

        if vol is not None:
            data = data.volume(vol, qpdata=True)

        if slice_loc is None:
            data_arr = data.raw()
            roi_arr = roi.raw()
        else:
            data_arr, _, _, _ = data.slice_data(slice_loc)
            roi_arr, _, _, _ = roi.slice_data(slice_loc)

        regions = []
        for region, name in roi.regions.items():
            # get data for a single label of the roi
            in_roi = data_arr[roi_arr == region]
            in_roi = in_roi[np.isfinite(in_roi)]
            if in_roi.size > 0:
                mean, med, std = np.mean(in_roi), np.median(in_roi), np.std(in_roi)
                mx, mn = np.max(in_roi), np.min(in_roi)
            else:
                mean, med, std, mx, mn = 0, 0, 0, 0, 0

            stat1['mean'].append(mean)
            stat1['median'].append(med)
            stat1['std'].append(std)
            stat1['max'].append(mx)
            stat1['min'].append(mn)
            regions.append(name)

            y, x = np.histogram(in_roi, bins=hist_bins, range=hist_range)
            hist1.append(y)
            hist1x.append(x)

        return stat1, regions, hist1, hist1x

class OverlayStatsProcess(DataStatisticsProcess):
    """
    For backwards compatibility
    """
    PROCESS_NAME = "OverlayStats"
    
class ExecProcess(Process):
    """
    Process which can execute arbitrary Python code
    """
    
    PROCESS_NAME = "Exec"
    
    def __init__(self, ivm, **kwargs):
        Process.__init__(self, ivm, **kwargs)
        self.model = QtGui.QStandardItemModel()

    def run(self, options):
        exec_globals = {'np': np, 'scipy' : scipy, 'ivm': self.ivm}

        # For general Numpy operations we will need a grid to put the
        # results back into. This is specified by the 'grid' option.
        # Note that all data is combined using their raw grids so these
        # must all match if the result is to work - that is the user's job!
        gridfrom = options.pop("grid", None)
        is_roi = options.pop("output-is-roi", False)

        if gridfrom is None:
            grid = self.ivm.main.grid
        else:
            grid = self.ivm.data[gridfrom].grid

        for name, data in self.ivm.data.items():
            exec_globals[name] = data.raw()

        for name in list(options.keys()):
            proc = options.pop(name)
            if name in ("exec", "_"):
                for code in proc:
                    try:
                        exec(code, exec_globals)
                    except:
                        raise QpException("'%s' is not valid Python code (Reason: %s)" % (code, sys.exc_info()[1]))
            else:
                try:
                    result = eval(proc, exec_globals)
                    self.ivm.add(result, grid=grid, name=name, roi=is_roi)
                except:
                    raise QpException("'%s' did not return valid data (Reason: %s)" % (proc, sys.exc_info()[1]))



"""
Quantiphyse - Analysis processes for Kmeans clustering

Copyright (c) 2013-2018 University of Oxford
"""

from __future__ import division, print_function, absolute_import

import time
import numpy as np
import sklearn.cluster as cl

from quantiphyse.data import NumpyData
from quantiphyse.processes import Process, normalisation, PCA
from quantiphyse.utils import QpException

class KMeansProcess(Process):
    """
    Clustering for a 4D volume
    """

    PROCESS_NAME = "KMeans"

    def __init__(self, ivm, **kwargs):
        Process.__init__(self, ivm, **kwargs)

    def run(self, options):
        data = self.get_data(options)
        roi = self.get_roi(options, data.grid)
        n_clusters = options.pop('n-clusters', 5)
        invert_roi = options.pop('invert-roi', False)
        output_name = options.pop('output-name', data.name + '_clusters')
        
        kmeans_data, mask = data.mask(roi, invert=invert_roi, output_flat=True, output_mask=True)
        start1 = time.time()

        if data.nvols > 1:
            # Do PCA reduction
            norm_data = options.pop('norm-data', True)
            norm_type = options.pop('norm-type', "sigenh")
            n_pca = options.pop('n-pca', 5)
            reduction = options.pop('reduction', 'pca')

            if reduction == "pca":
                self.log("Using PCA dimensionality reduction")
                pca = PCA(n_components=n_pca, norm_input=True, norm_type=norm_type,
                          norm_modes=norm_data)
                kmeans_data = pca.get_training_features(kmeans_data)
            else:
                raise QpException("Unknown reduction method: %s" % reduction)
        else:
            kmeans_data = kmeans_data[:, np.newaxis]

        kmeans = cl.KMeans(init='k-means++', n_clusters=n_clusters, n_init=10, n_jobs=1)
        kmeans.fit(kmeans_data)
        
        self.log("Elapsed time: %s" % (time.time() - start1))

        label_image = np.zeros(data.grid.shape, dtype=np.int)
        label_image[mask] = kmeans.labels_ + 1
        self.ivm.add(NumpyData(label_image, grid=data.grid, name=output_name, roi=True), make_current=True)

class MeanValuesProcess(Process):
    """
    Create new data set by replacing voxel values with mean within each ROI region
    """
    PROCESS_NAME = "MeanValues"
    
    def __init__(self, ivm, **kwargs):
        Process.__init__(self, ivm, **kwargs)

    def run(self, options):
        data = self.get_data(options)
        roi = self.get_roi(options, data.grid)
        output_name = options.pop('output-name', data.name + "_means")

        in_data = data.raw()
        out_data = np.zeros(in_data.shape)
        for region in roi.regions:
            if data.ndim > 3:
                out_data[roi.raw() == region] = np.mean(in_data[roi.raw() == region], axis=0)
            else:
                out_data[roi.raw() == region] = np.mean(in_data[roi.raw() == region])

        self.ivm.add(NumpyData(out_data, grid=data.grid, name=output_name), make_current=True)



"""
Quantiphyse - Widgets allowing the user to build simple ROIs

A new ROI is based on the grid of an existing data set - chosen by the user.

Copyright (c) 2013-2018 University of Oxford
"""

from __future__ import division, unicode_literals, absolute_import, print_function

import collections

import numpy as np

try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets

from quantiphyse.data import NumpyData
from quantiphyse.gui.options import OptionBox, DataOption, NumericOption, TextOption
from quantiphyse.gui.widgets import QpWidget, TitleWidget
from quantiphyse.gui.pickers import PickMode
from quantiphyse.utils import get_icon

from .tools import CrosshairsTool, PenTool, WalkerTool, PainterTool, EraserTool, RectTool, EllipseTool, PolygonTool, PickTool, BucketTool

DESC = """
Widget for creating test ROIs and basic manual segmentation
"""

TOOLS = [CrosshairsTool(), PenTool(), PainterTool(), EraserTool(), RectTool(), EllipseTool(), PolygonTool(), PickTool(), BucketTool()]

class RoiBuilderWidget(QpWidget):
    """
    Widget for building ROIs
    """
    ADD = 1
    ERASE = 2
    MASK = 3

    def __init__(self, **kwargs):
        super(RoiBuilderWidget, self).__init__(name="ROI Builder", icon="roi_builder", desc=DESC,  
                                               group="ROIs", **kwargs)
        self._history = collections.deque(maxlen=10)
        self._tool = None
        self.grid = None
        self.roi = None
        self.roiname = None

    def init_ui(self):
        layout = QtGui.QVBoxLayout()
        
        title = TitleWidget(self, help="roibuilder", batch_btn=False)
        layout.addWidget(title)

        self.options = OptionBox("Options")
        btn = QtGui.QPushButton("New")
        btn.clicked.connect(self._new_roi)
        self.options.add("ROI", DataOption(self.ivm, rois=True, data=False), btn, key="roi")
        self.options.add("Current label", NumericOption(minval=1, slider=False, intonly=True), key="label")
        self.options.add("Label description", TextOption(), key="label_text")
        self.options.option("roi").sig_changed.connect(self._roi_changed)
        self.options.option("label").sig_changed.connect(self._label_changed)
        self.options.option("label_text").sig_changed.connect(self._label_text_changed)
        layout.addWidget(self.options)
        
        # Add toolbox buttons in a grid
        hbox = QtGui.QHBoxLayout()
        self._toolbox = QtGui.QGroupBox()
        self._toolbox.setTitle("Toolbox")
        self.tools_grid = QtGui.QGridLayout()
        self._toolbox.setLayout(self.tools_grid)

        x, y, cols = 0, 0, 4
        for tool in TOOLS:
            self._add_tool(tool, y, x)
            x += 1
            if x == cols:
                y += 1
                x = 0

        self._undo_btn = QtGui.QPushButton()
        self._undo_btn.clicked.connect(self.undo)
        self._undo_btn.setEnabled(False)
        undo_icon = QtGui.QIcon(get_icon("undo"))
        self._undo_btn.setIcon(undo_icon)
        self._undo_btn.setToolTip("Undo last action")
        self._undo_btn.setFixedSize(32, 32)
        self.tools_grid.addWidget(self._undo_btn, y, x)

        hbox.addWidget(self._toolbox)
        self._toolbox.setEnabled(False)
        hbox.addStretch(1)
        layout.addLayout(hbox)

        # Tool options box - initially invisible
        hbox = QtGui.QHBoxLayout()
        self._tool_options = QtGui.QGroupBox()
        self._tool_options.setVisible(False)

        hbox.addWidget(self._tool_options)
        hbox.addStretch(1)
        layout.addLayout(hbox)

        layout.addStretch(1)
        self.setLayout(layout)

    def activate(self):
        self._roi_changed()
        self.ivl.set_picker(PickMode.SINGLE)
        if self._tool is not None:
            self._tool.selected()

    def deactivate(self):
        self.ivl.set_picker(PickMode.SINGLE)
        if self._tool is not None:
            self._tool.deselected()

    def modify(self, vol=None, slice2d=None, points=None, mode=None):
        """
        Make a change to the ROI we are building

        :param roi_new: Numpy array containing the updated ROI data on the current base grid
        :param vol: 3D 3D selection specified as binary Numpy array same shape as ROI
        :param slice2d: 2D slice selection specified as tuple of (2D binary Numpy array, axis index, position)
        :param points: Selection specified as list of 3D co-ordinates
        :param mode: ``ADD`` to add selection to the ROI (using current label), ``ERASE`` to erase
                     selection from the ROI (set to 0), ``MASK`` to preserve ROI labels in
                     selection but zero everything outside selection
        """
        label = self.options.option("label").value
        self.debug("label=%i", label)

        # For undo functionality: selection is an object specifying which
        # points or ROI region were selected, data_orig is a corresponding
        # object which contains the data before the operation occurred.
        selection, data_orig = None, None

        if points is not None:
            selection = points
            data_orig = [self.roidata[point[0], point[1], point[2]] for point in points]
            for point in points:
                if mode == self.ADD:
                    self.roidata[point[0], point[1], point[2]] = label
                elif mode == self.ERASE:
                    self.roidata[point[0], point[1], point[2]] = 0
                else:
                    raise ValueError("Invalid mode: %i" % mode)
        else:
            # change_subset is a Numpy selection of the points to be
            # affected, data_new is a corresponding binary array identifying
            # the selected points in this subset to modify.
            if vol is not None:
                # Selection left as None to indicate whole volume
                selected_points = vol
                change_subset = self.roidata
            elif slice2d is not None:
                selected_points, axis, pos = slice2d
                slices = [slice(None)] * 3
                slices[axis] = pos
                change_subset = self.roidata[slices]
                # Selection is the axis index and position
                selection = (axis, pos)
            else:
                raise ValueError("Neither volume nor slice nor points provided")

            data_orig = np.copy(change_subset)
            if mode == self.ADD:
                self.debug("Adding: %i", np.count_nonzero(selected_points))
                change_subset[selected_points > 0] = label
            elif mode == self.ERASE:
                self.debug("Erasing: %i", np.count_nonzero(selected_points))
                change_subset[selected_points > 0] = 0
            elif mode == self.MASK:
                self.debug("Masking: %i", np.count_nonzero(selected_points))
                change_subset[selected_points == 0] = 0
            else:
                raise ValueError("Invalid mode: %i" % mode)
        
        # Save the previous state of the data in the history list
        self._history.append((selection, data_orig))
        self._undo_btn.setEnabled(True)
        
        # Update the ROI - note that the regions may have been affected so make
        # sure they are regenerated
        self._update_regions()
        self.ivl.redraw()
        self.debug("Now have %i nonzero", np.count_nonzero(self.roidata))

    def _update_regions(self):
        """
        ROI regions may have been created or added so regenerate them but put 
        back existing label names
        """
        current_regions = self.ivm.data[self.roiname].metadata.pop("roi_regions", {})
        new_regions = self.ivm.data[self.roiname].regions
        for label, desc in current_regions.items():
            if label in new_regions:
                new_regions[label] = desc
        
    def undo(self):
        """
        Undo the last change
        """
        self.debug("ROI undo: %i", len(self._history))
        if not self._history: 
            return

        selection, data_orig = self._history.pop()

        # For selection, None indicates whole volume, tuple indicates
        # an (axis, pos) slice, otherwise we have a sequence of points
        if selection is None:
            self.roidata[:] = data_orig
        elif isinstance(selection, tuple):
            axis, pos = selection
            slices = [slice(None)] * 3
            slices[axis] = pos
            self.roidata[slices] = data_orig
        else:
            for point, orig_value in zip(selection, data_orig):
                self.roidata[point[0], point[1], point[2]] = orig_value

        self._update_regions()
        self.ivl.redraw()
        self.debug("Now have %i nonzero", np.count_nonzero(self.roidata))
        self._undo_btn.setEnabled(len(self._history) > 0)
      
    def _label_changed(self):
        self.debug("Label changed")
        roi = self.ivm.data.get(self.options.option("roi").value, None)
        if roi is not None:
            label = self.options.option("label").value
            self.debug(label)
            regions = roi.regions
            if label in regions:
                self.options.option("label_text").value = regions[label]
            else:
                self.options.option("label_text").value = "Region %i" % label
        
    def _label_text_changed(self):
        self.debug("Label text changed")
        roi = self.ivm.data.get(self.options.option("roi").value, None)
        if roi is not None:
            label = self.options.option("label").value
            label_text = self.options.option("label_text").value
            self.debug(label)
            self.debug(label_text)
            regions = roi.regions
            regions[label] = label_text

    def _roi_changed(self):
        roi = self.ivm.data.get(self.options.option("roi").value, None)
        self._toolbox.setEnabled(roi is not None)
        if roi is not None:
            # FIXME this will only work if ROI is NumpyData. Otherwise we are
            # manipulating a numpy array which may just be a proxy for the file
            # storage.
            regions = roi.regions
            current_label = self.options.option("label").value
            if self.roiname != roi.name or current_label not in regions.keys():
                self.options.option("label").value = min(list(regions.keys()) + [1, ])
            self.roiname = roi.name
            self.grid = roi.grid
            self.roidata = roi.raw()

    def _new_roi(self):
        dialog = QtGui.QDialog(self)
        dialog.setWindowTitle("New ROI")
        vbox = QtGui.QVBoxLayout()
        dialog.setLayout(vbox)

        optbox = OptionBox()
        optbox.add("ROI name", TextOption(), key="name")
        optbox.add("Data space from", DataOption(self.ivm), key="grid")
        vbox.addWidget(optbox)

        buttons = QtGui.QDialogButtonBox(QtGui.QDialogButtonBox.Ok | QtGui.QDialogButtonBox.Cancel)
        buttons.accepted.connect(dialog.accept)
        buttons.rejected.connect(dialog.reject)
        vbox.addWidget(buttons)
        
        ok = dialog.exec_()
        if ok:
            roiname = optbox.option("name").value
            grid = self.ivm.data[optbox.option("grid").value].grid
            roidata = np.zeros(grid.shape, dtype=np.int)
            self.ivm.add(NumpyData(roidata, grid=grid, roi=True, name=roiname), make_current=True)

            # Throw away old history. FIXME is this right, should we keep existing data and history?
            # Also should we cache old history in case we go back to this ROI?
            self._history = []
            self._undo_btn.setEnabled(False)
            self.options.option("roi").value = roiname

    def _tool_selected(self, tool):
        def _select():
            if self._tool is not None:
                self._tool.btn.setStyleSheet("")
                self._tool.deselected()
            
            self._tool = tool
            self._tool.btn.setStyleSheet("border: 2px solid QLinearGradient( x1: 0, y1: 0, x2: 0, y2: 1, stop: 0 #ffa02f, stop: 1 #d7801a);")
            # Replace the old tool options with the new one. Need to reparent the
            # existing layout to a temporary widget which will then get deleted 
            QtGui.QWidget().setLayout(self._tool_options.layout())
            self._tool_options.setLayout(self._tool.interface())
            self._tool_options.setTitle(tool.name)
            self._tool_options.setVisible(True)
            self._tool.selected()

        return _select
    
    def _add_tool(self, tool, x, y):
        tool.ivm = self.ivm
        tool.ivl = self.ivl
        tool.builder = self
        btn = QtGui.QPushButton()
        btn.setIcon(QtGui.QIcon(get_icon(tool.name.lower())))
        btn.setToolTip(tool.tooltip)
        btn.setFixedSize(32, 32)
        btn.clicked.connect(self._tool_selected(tool))
        tool.btn = btn
        self.tools_grid.addWidget(btn, x, y)



"""
Quantiphyse - Processes for the data manipulation package

Copyright (c) 2013-2018 University of Oxford
"""

from quantiphyse.utils import QpException

from quantiphyse.processes import Process

class ResampleProcess(Process):
    """ 
    Resample data 
    """

    PROCESS_NAME = "Resample"
    
    def run(self, options):
        data = self.get_data(options)
        if data.roi: 
            default_order=0
        else:
            default_order=1
        order = options.pop("order", default_order)
        output_name = options.pop("output-name", "%s_res" % data.name)
        grid_data = options.pop("grid", None)

        if grid_data is None:
            raise QpException("Must provide 'grid' option to specify data item to get target grid from")
        elif grid_data not in self.ivm.data:
            raise QpException("Data item '%s' not found" % grid_data)
        
        grid = self.ivm.data[grid_data].grid
        output_data = data.resample(grid, order=order)
        output_data.name = output_name
        self.ivm.add(output_data, make_current=True, roi=data.roi and order == 0)



import os

import pandas as pd

from quantiphyse.processes import Process
from quantiphyse.test import ProcessTest
from quantiphyse.data.extras import DataFrameExtra

class IoProcessTest(ProcessTest):

    def testSaveAllExcept(self):
        yaml = """
  - SaveAllExcept:
        data_3d:
        data_4d_moving:
"""
        self.run_yaml(yaml)
        self.assertEqual(self.status, Process.SUCCEEDED)
        self.assertTrue(os.path.exists(os.path.join(self.output_dir, "case", "data_4d.nii")))
        self.assertTrue(os.path.exists(os.path.join(self.output_dir, "case", "mask.nii")))
        self.assertFalse(os.path.exists(os.path.join(self.output_dir, "case", "data_4d_moving.nii")))
        self.assertFalse(os.path.exists(os.path.join(self.output_dir, "case", "data_3d.nii")))

    def testSaveAll(self):
        yaml = """
  - SaveAllExcept:
"""
        self.run_yaml(yaml)
        self.assertEqual(self.status, Process.SUCCEEDED)
        self.assertTrue(os.path.exists(os.path.join(self.output_dir, "case", "data_4d.nii")))
        self.assertTrue(os.path.exists(os.path.join(self.output_dir, "case", "mask.nii")))
        self.assertTrue(os.path.exists(os.path.join(self.output_dir, "case", "data_4d_moving.nii")))
        self.assertTrue(os.path.exists(os.path.join(self.output_dir, "case", "data_3d.nii")))

    def testSave(self):
        yaml = """
  - Save:
        data_3d:
        data_4d: data_multivol
"""
        self.run_yaml(yaml)
        self.assertEqual(self.status, Process.SUCCEEDED)
        self.assertTrue(os.path.exists(os.path.join(self.output_dir, "case", "data_multivol.nii")))
        self.assertTrue(os.path.exists(os.path.join(self.output_dir, "case", "data_3d.nii")))
        self.assertTrue("data_3d" in self.ivm.data)
        self.assertTrue("data_4d" in self.ivm.data)

    def testSaveDelete(self):
        yaml = """
  - SaveAndDelete:
        data_3d:
        data_4d: data_multivol
"""
        self.run_yaml(yaml)
        self.assertEqual(self.status, Process.SUCCEEDED)
        self.assertTrue(os.path.exists(os.path.join(self.output_dir, "case", "data_multivol.nii")))
        self.assertTrue(os.path.exists(os.path.join(self.output_dir, "case", "data_3d.nii")))
        self.assertFalse("data_3d" in self.ivm.data)
        self.assertFalse("data_4d" in self.ivm.data)
        
    def testSaveExtras(self):
        yaml = """
  - SaveExtras:
        my_extra: saved_file.mat
"""
        df = pd.DataFrame({'one': [1., 2., 3., 4.], 'two': [4., 3., 2., 1.]})
        self.ivm.add_extra("my_extra", DataFrameExtra("my_extra", df))
        self.run_yaml(yaml)
        self.assertEqual(self.status, Process.SUCCEEDED)
        self.assertTrue(os.path.exists(os.path.join(self.output_dir, "case", "saved_file.mat")))
        



"""
Quantiphyse - Implements the batch processing system

The ``Script`` class defines a processing pipeline. It consists
of a series of ``Process`` objects and, optionally, a series of ``BatchCase``
objects, each of which describes a separate execution of the same process
pipeline. It is also in itself a ``Process``.

``Script`` can be used as a means of defining a macro process but is also
used as the basis of the batch processing system. The pipeline can be
defined programatically or via a YAML document.

``BatchScript`` is a subclass of ``Script`` which adds human-readable 
output suitable for command line batch execution.

Copyright (c) 2013-2018 University of Oxford
"""

import sys
import os
import os.path
import traceback
import time
import collections
import logging

import six
import yaml
import numpy as np

try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets

from quantiphyse.processes import Process
from quantiphyse.processes.io import *
from quantiphyse.processes.misc import *
from quantiphyse.utils.logger import set_base_log_level
from quantiphyse.data import ImageVolumeManagement, load, save

from . import get_plugins, ifnone
from .exceptions import QpException

# Default basic processes - all others are imported from packages
BASIC_PROCESSES = {
    "RenameData"   : RenameProcess,
    "RenameRoi"   : RenameProcess,
    "Rename"   : RenameProcess,
    "Delete"   : DeleteProcess,
    "RoiCleanup" : RoiCleanupProcess,
    "Load" : LoadProcess,
    "Save" : SaveProcess,
    "SaveAllExcept" : SaveAllExceptProcess,
    "SaveAndDelete" : SaveDeleteProcess,
    "LoadData" : LoadDataProcess,
    "LoadRois" : LoadRoisProcess,
    "SaveArtifacts" : SaveArtifactsProcess,
    "SaveExtras" : SaveArtifactsProcess
}

def to_yaml(processes, indent=""):
    """
    Turn a process list into YAML

    The reason for using this function instead of the built-in PyYAML
    dump functions is to get the flow we want and also handle Numpy 
    arrays. In particular, we want lists to be inline but dictionaries
    to be in block format.

    This only supports a small subset of YAML types:
    
     - The processes must either be a list of dictionaries, or a single dictionary.
     - Dictionary keys must be strings
     - Dictionary values must be strings, numbers, lists, Numpy arrays or dictionaries
    
    Anything else will throw a ValueError. While the list of supported types might
    increase in the future, part of the intention is to constrain the types that
    processes can use in for options, so it will not change much.

    This does not affect the parsing of YAML code which uses PyYAML and can use
    any supported YAML code.
    """
    def _dict_to_yaml(stream, valdict, indent="", prefix=""):
        for key, value in valdict.items():
            if not isinstance(key, six.string_types):
                try:
                    key = str(int(key))
                except ValueError:
                    raise ValueError("Keys must be strings or ints")

            if value is None:
                continue

            stream.write("%s%s: " % (prefix, key))
            if isinstance(value, six.string_types):
                stream.write("%s\n" % value)
            elif isinstance(value, (int, float, list, np.floating, np.integer)):
                stream.write("%s\n" % str(value))
            elif isinstance(value, np.ndarray):
                stream.write("%s\n" % str(value.tolist()))
            elif isinstance(value, collections.Sequence):
                stream.write("%s\n" % str(list(value)))
            elif isinstance(value, collections.Mapping):
                stream.write("\n")
                _dict_to_yaml(stream, value, indent + "  ", prefix=indent + "  ")
            else:
                raise ValueError("Unsupported option value type: %s" % type(value))            

    if isinstance(processes, dict):
        processes = [processes,]

    yaml_str = six.StringIO()
    for process in processes:
        _dict_to_yaml(yaml_str, process, indent=indent + "    ", prefix="  - ")
        yaml_str.write("\n")
    return yaml_str.getvalue()

class Script(Process):
    """
    A processing script. It consists of three types of information:

     - Generic options (e.g. debug mode)
     - A single pipeline of processing steps
     - Optional list of BatchCase objects to apply these steps to

    A batch script can be run on a specified IVM, or it can be
    run on its cases. In this case a new IVM is created for
    each case
    """

    PROCESS_NAME = "Script"

    IGNORE = 1
    NEXT_CASE = 2
    FAIL = 3

    sig_start_case = QtCore.Signal(object)
    sig_done_case = QtCore.Signal(object)
    sig_start_process = QtCore.Signal(object, dict)
    sig_process_progress = QtCore.Signal(float)
    sig_done_process = QtCore.Signal(object, dict)

    def __init__(self, ivm=None, **kwargs):
        """
        fname: File name containing YAML code to load from
        code: YAML code as a string
        yamlroot: Parsed YAML code as Python objects
        """
        super(Script, self).__init__(ivm, **kwargs)
        
        self._current_ivm = None
        self._current_process = None
        self._current_params = None
        self._process_num = 0
        self._process_start = None
        self._current_case = None
        self._case_num = 0
        self._pipeline = []
        self._cases = []
        self._generic_params = {}
        self._error_action = kwargs.get("error_action", Script.IGNORE)
        self._embed_log = kwargs.get("embed_log", False)
        self._output_items = []

        # Find all the process implementations
        self.known_processes = dict(BASIC_PROCESSES)
        plugin_processes = get_plugins("processes")
        for process in plugin_processes:
            self.known_processes[process.PROCESS_NAME] = process

    def run(self, options):
        """
        Run the script

        The pipeline is created first, either from supplied
        YAML code, a filename or a set of parsed objects

        If no ``Case`` are included, a single default ``Case``
        is created. The completed pipeline is then run on all
        cases.

        Runs are asynchronous - initially we start the first
        process of the first case, and connect to the 'process finished'
        signal. When the slot is called, we start the next process, 
        or the next case as required. So the ``run()`` method returns
        as soon as the first process is started. 
        """
        if "parsed-yaml" in options:
            root = dict(options.pop("parsed-yaml"))
        elif "yaml" in options:
            root = yaml.load(options.pop("yaml"))
        elif "yaml-file" in options:
            with open(options.pop("yaml-file"), "r") as yaml_file:
                root = yaml.load(yaml_file)
        else:
            raise RuntimeError("Neither filename nor YAML code provided")

        if root is None: 
            # Handle special case of empty content
            root = {}

        # Can set mode=check to just validate the YAML
        self._load_yaml(root)
        self.debug(self._pipeline)
        self._output_items = []
        mode = options.pop("mode", "run")
        if mode == "run":
            self.status = Process.RUNNING
            self._case_num = 0
            self._next_case()
        elif mode != "check":
            raise QpException("Unknown mode: %s" % mode)

    def cancel(self):
        if self._current_process is not None:
            self._current_process.cancel()
    
    def _load_yaml(self, root=None):
        """
        Load YAML content
        """
        self._pipeline = []
        for process in root.pop("Processing", []):
            name = list(process.keys())[0]
            proc = self.known_processes.get(name, None)
            params = process[name]
            if params is None: params = {}

            if proc is None:
                raise QpException("Unknown process: %s" % name)
            else:
                params["id"] = params.get("id", name)
                params["__impl"] = proc
                self._pipeline.append(params)

        # Cases can be expressed as list or dict
        self._cases = []
        yaml_cases = root.pop("Cases", [])
        if isinstance(yaml_cases, dict):
            for case_id in sorted(yaml_cases.keys()):
                self._cases.append(Case(case_id, yaml_cases[case_id]))
        else:
            for case in yaml_cases:
                case_id = list(case.keys())[0]
                self._cases.append(Case(case_id, case.get(case_id, {})))
        
        # Create default case if we have not been specified any
        if not self._cases:
            self._cases.append(Case("case", {}))
        # After removing processes and cases, remainder is the generic options
        self._generic_params = root
    
    def _next_case(self):
        if self.status != self.RUNNING:
            return
        
        if self._case_num < len(self._cases):
            case = self._cases[self._case_num]
            self._case_num += 1
            self.sig_start_case.emit(case)
            self.debug("Starting case %s", case.case_id)
            self._start_case(case)
        else:
            self.debug("All cases complete")
            self.status = Process.SUCCEEDED
            self._complete()

    def _start_case(self, case):
        if self.ivm is not None:
            self._current_ivm = self.ivm
        else:
            self._current_ivm = ImageVolumeManagement()
        self._current_case = case
        self._process_num = 0
        self._next_process()

    def _next_process(self):
        if self.status != self.RUNNING:
            return
        
        if self._process_num < len(self._pipeline):
            process = self._pipeline[self._process_num]
            self._process_num += 1
            self._start_process(process)
        else:
            self.debug("All processes complete")
            if len(self._cases) > 1:
                self.log("CASE COMPLETE\n")
            self.sig_done_case.emit(self._current_case)
            self._next_case()

    def _start_process(self, proc_params):
        # Make copy so process does not mess up shared config
        proc_params = dict(proc_params)
        generic_params = dict(self._generic_params)

        # Override values which are defined in the individual case
        if self._current_case is not None:
            case_params = dict(self._current_case.params)
            override = case_params.pop(proc_params["id"], {})
            proc_params.update(override)
            generic_params.update(case_params)
            # OutputId defaults to the case ID if not specified
            if "OutputId" not in generic_params:
                generic_params["OutputId"] = self._current_case.case_id

        # Set debug level for this individual process based on whether logging
        # was enabled generically, for this case, and for this process
        if "--debug" in sys.argv or proc_params.get("Debug", generic_params.get("Debug", False)):
            set_base_log_level(logging.DEBUG)
        else:
            set_base_log_level(logging.WARN)

        try:
            outdir = os.path.abspath(os.path.join(ifnone(generic_params.get("OutputFolder", ""), ""), 
                                                  ifnone(generic_params.get("OutputId", ""), ""),
                                                  ifnone(generic_params.get("OutputSubFolder", ""), "")))
            indir = os.path.abspath(os.path.join(ifnone(generic_params.get("InputFolder", generic_params.get("Folder", "")), ""), 
                                                 ifnone(generic_params.get("InputId", ""), ""),
                                                 ifnone(generic_params.get("InputSubFolder", ""), "")))
            
            proc_id = proc_params.pop("id")
            process = proc_params.pop("__impl")(self._current_ivm, indir=indir, outdir=outdir, proc_id=proc_id)
            
            self._current_process = process
            self._current_params = proc_params
            process.sig_finished.connect(self._process_finished)
            process.sig_progress.connect(self._process_progress)
            process.sig_log.connect(self._process_log)
            
            self._process_start = time.time()
            if len(self._pipeline) > 1:
                self.log("Running %s\n\n" % process.proc_id)
            for key, value in proc_params.items():
                self.debug("      %s=%s" % (key, str(value)))

            self.sig_start_process.emit(process, dict(proc_params))
            process.execute(proc_params)
        
        except Exception as exc:
            # Could not create process - treat as process failure
            self._process_finished(Process.FAILED, "Process failed to start: " + str(exc), exc)
        finally:
            pass

    def _process_finished(self, status, log, exception):
        self.debug("Process finished: %s", self._current_process.proc_id)
        self._current_process.sig_finished.disconnect(self._process_finished)
        self._current_process.sig_progress.disconnect(self._process_progress)
        self._current_process.sig_log.disconnect(self._process_log)
        if self.status != self.RUNNING:
            return

        end = time.time()
        self.sig_done_process.emit(self._current_process, dict(self._current_params))
        
        if status == Process.SUCCEEDED:
            if len(self._pipeline) > 1:
                self.log("\nDONE (%.1fs)\n" % (end - self._process_start))
            self._output_items.extend(self._current_process.output_data_items())
            self._next_process()
        else:
            self.log("\nFAILED: %i\n" % status)
            if self._error_action == Script.IGNORE:
                self.debug("Process failed - ignoring")
                self._next_process()
            elif self._error_action == Script.FAIL:
                self.debug("Process failed - stopping script")
                self.status = status
                self.exception = exception
                self._current_process = None
                self._current_params = None
                self._complete()
            elif self._error_action == Script.NEXT_CASE:
                self.debug("Process failed - going to next case")
                self.log("CASE FAILED\n")
                self.sig_done_case.emit(self._current_case)
                self._next_case()

    def _process_progress(self, complete):
        self.sig_process_progress.emit(complete)
        script_complete = ((self._case_num-1)*len(self._pipeline) + 
                           (self._process_num - 1 + complete)) / (len(self._pipeline)*len(self._cases))
        self.sig_progress.emit(script_complete)

    def _process_log(self, msg):
        self.log(msg)
        
    def output_data_items(self):
        return self._output_items

class Case(object):
    """
    An individual case (e.g. patient scan) which a processing pipeline is applied to
    """
    def __init__(self, case_id, params):
        self.case_id = case_id
        if params is None:
            params = {}
        self.params = params
        # This would break compatibility so not for now
        #self.params["InputId"] = self.params.get("InputId", self.case_id)

class BatchScript(Script):
    """
    A Script which sends human readable output to a log stream. It also
    saves the logs of the processes to a file in the output folder.

    This is used as the runner for batch scripts started from the console
    or from the ``BatchBuilder`` widget.
    """
    def __init__(self, ivm=None, stdout=sys.stdout, **kwargs):
        Script.__init__(self, ivm, **kwargs)
        self.stdout = stdout
        self.start = None
        self._quit_on_exit = kwargs.get("quit_on_exit", True)

        self.sig_start_case.connect(self._log_start_case)
        self.sig_done_case.connect(self._log_done_case)
        self.sig_start_process.connect(self._log_start_process)
        self.sig_process_progress.connect(self._log_process_progress)
        self.sig_done_process.connect(self._log_done_process)
        self.sig_progress.connect(self._log_progress)
        self.sig_finished.connect(self._log_done_script)

    def _log_start_case(self, case):
        self.stdout.write("Processing case: %s\n" % case.case_id)

    def _log_done_case(self, case):
        pass

    def _log_start_process(self, process, params):
        self.start = time.time()
        self.stdout.write("  - Running %s...  0%%" % process.proc_id)
        for key, value in params.items():
            self.debug("      %s=%s" % (key, str(value)))
                
    def _log_done_process(self, process, params):
        if process.status == Process.SUCCEEDED:
            end = time.time()
            self.stdout.write(" DONE (%.1fs)\n" % (end - self.start))
            fname = os.path.join(process.outdir, "%s.log" % process.proc_id)
            self._save_text(process.get_log(), fname)
            if params:
                self.warn("Unused parameters")
                for key, val in params.items():
                    self.warn("%s=%s", str(key), str(val))
        else:
            self.stdout.write(" FAILED: %i\n" % process.status)
            self.warn(str(process.exception))
            self.debug(traceback.format_exc(process.exception))

    def _log_progress(self, complete):
        #self.stdout.write("%i%%\n" % int(100*complete))
        pass

    def _log_process_progress(self, complete):
        percent = int(100*complete)
        self.stdout.write("\b\b\b\b%3i%%" % percent)
        self.stdout.flush()

    def _log_done_script(self):
        self.stdout.write("Script finished\n")
        if self._quit_on_exit:
            QtCore.QCoreApplication.instance().quit()

    def _save_text(self, text, fname, ext="txt"):
        if text:
            if "." not in fname: fname = "%s.%s" % (fname, ext)
            dirname = os.path.dirname(fname)
            if not os.path.exists(dirname): os.makedirs(dirname)
            with open(fname, "w") as text_file:
                text_file.write(text)



"""
Quantiphyse - Radial profile widget

Copyright (c) 2013-2018 University of Oxford
"""
import csv

import six

try:
    from PySide import QtGui, QtCore, QtGui as QtWidgets
except ImportError:
    from PySide2 import QtGui, QtCore, QtWidgets
  
from quantiphyse.gui.widgets import QpWidget, TitleWidget
from quantiphyse.gui.plot import Plot
from quantiphyse.gui.options import OptionBox, DataOption, NumericOption, BoolOption

from .process import RadialProfileProcess
    
class RadialProfileWidget(QpWidget):
    """
    Widget which displays radial profile of data
    """
    def __init__(self, **kwargs):
        super(RadialProfileWidget, self).__init__(name="Radial profile", 
                                                  icon="rp",
                                                  desc="Display radial profile of data", 
                                                  group="Visualisation", **kwargs)
        self._updating = False

    def init_ui(self):
        vbox = QtGui.QVBoxLayout()
        self.setLayout(vbox)

        title = TitleWidget(self)
        vbox.addWidget(title)

        self.options = OptionBox("Options")
        self.options.add("Data", DataOption(self.ivm, multi=True), key="data")
        self.options.add("Within ROI", DataOption(self.ivm, data=False, rois=True, none_option=True), key="roi")
        self.options.add("All volumes", BoolOption(default=False), key="allvols")
        self.options.add("Number of bins", NumericOption(minval=5, maxval=250, default=50, intonly=True), key="bins")
        vbox.addWidget(self.options)

        self.plot = Plot(qpo=None, parent=self, title="Radial profile", display_mode=False)
        self.plot.set_xlabel("Distance (mm)")
        self.plot.set_ylabel("Mean data value")
        vbox.addWidget(self.plot)

        vbox.addStretch(1)

    def activate(self):
        self.ivm.sig_all_data.connect(self._data_changed)
        self.ivl.sig_focus_changed.connect(self._update)
        self.options.option("data").sig_changed.connect(self._data_changed)
        self.options.sig_changed.connect(self._update)
        self._data_changed()

    def deactivate(self):
        self.ivm.sig_all_data.disconnect(self._data_changed)
        self.ivl.sig_focus_changed.disconnect(self._update)
        self.options.option("data").sig_changed.disconnect(self._data_changed)
        self.options.sig_changed.disconnect(self._update)

    def processes(self):
        opts = self.options.values()
        opts["centre"] = self.ivl.focus()
        if opts.pop("allvols", False):
            opts["centre"] = opts["centre"][:3]

        return {
            "RadialProfile" : opts
        }

    def _data_changed(self):
        if self._updating: return
        self._updating = True
        try:
            data_names = self.options.option("data").value
            multivol = False
            for data_name in data_names:
                multivol = multivol or self.ivm.data[data_name].nvols > 1
            self.options.set_visible("allvols", multivol)
            self._update()
        finally:
            self._updating = False

    def _update(self):
        process = RadialProfileProcess(self.ivm)
        process.execute(self.processes()["RadialProfile"])
        self._update_plot()

    def _update_plot(self):
        self.plot.clear()
        rp = self.ivm.extras.get("radial-profile", None)
        if rp is not None:
            xvalues = rp.df.index
            for idx, name in enumerate(rp.df.columns):
                yvalues = rp.df.values[:,idx]
                self.plot.add_line(yvalues, name=name, xvalues=xvalues)
